/*
 * @(#)DrawApplication.java 5.1
 *
 */

package CH.ifa.draw.application;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;
import CH.ifa.draw.framework.*;
import CH.ifa.draw.standard.*;
import CH.ifa.draw.figures.*;
import CH.ifa.draw.contrib.*;
import CH.ifa.draw.util.*;

// for class DrawingLoadServer:
import java.io.*;
import java.net.*;
import java.util.*;

/**
 * DrawApplication defines a standard presentation for
 * standalone drawing editors. The presentation is
 * customized in subclasses.
 * The application is started as follows:
 * <pre>
 * public static void main(String[] args) {
 *     MyDrawApp window = new MyDrawApp();
 * }
 * </pre>
 */

public  class DrawApplication
        implements DrawingEditor, PaletteListener {

    private static int windowCount = 0;

    protected Frame             menuFrame, toolFrame;
    private final String        defaultDrawExtension=".draw";
    public String               drawExtension;
    private boolean             floating;
    private Vector              fDrawings = new Vector();
    private Tool                fTool;
    private Iconkit             fIconkit;
    private Menu                fDrawingsMenu;
    private TextField           fStatusLine;
    private Vector              fViewFrames = new Vector();
    private DrawingViewFrame    fViewFrame;
    private boolean             fToolButtonSticky=false;
    private ToolButton          fDefaultToolButton;
    private ToolButton          fSelectedToolButton = null;
    private ToolButton          fConnTextTB, fTextTB;

    // the image resource path
    private static final String fgDrawPath = "/CH/ifa/draw/";
    public static final String IMAGES = fgDrawPath+"images/";

    /**
     * The index of the file menu in the menu bar.
     */
    public static final int    FILE_MENU = 0;
    /**
     * The index of the edit menu in the menu bar.
     */
    public static final int    EDIT_MENU = 1;
    /**
     * The index of the alignment menu in the menu bar.
     */
    public static final int    ALIGNMENT_MENU = 2;
    /**
     * The index of the attributes menu in the menu bar.
     */
    public static final int    ATTRIBUTES_MENU = 3;

    /**
     * The index of the drawings menu in the menu bar.
     */
    public static final int    DRAWINGS_MENU = 5;

    private String fFilename = "";

    /**
     * Constructs a drawing window with a default title.
     */
    public DrawApplication() {
       this("JHotDraw");
    }

    /**
     * Constructs a drawing window with the given title.
     */
    public DrawApplication(String title) {
       this(title, null, new String[0], null);
    }

    public DrawApplication(String title, String[] drawings) {
       this(title,null,drawings,null);
    }

    public DrawApplication(String title, String drawExtension, String[] drawings, String icon) {

        DrawingLoadServer dls=null;
        String serverPortStr=System.getProperty("de.renew.loadServerPort");
        if (serverPortStr!=null) {
          try {
             int serverPort=Integer.valueOf(serverPortStr).intValue();
             try {
                dls=new DrawingLoadServer(this,serverPort);
             } catch (IOException ioe) {
                System.out.println("Server Socket is occupied!");
                System.exit(-1);
             }
          } catch (Exception e) {
             System.out.println("No Drawing Load Server: "+e);
          }
        } /* endif */

        menuFrame = new Frame(title);
        fStatusLine = createStatusLine();
        menuFrame.add("South", fStatusLine);

//        toolFrame = new Frame("Tools");
//        toolFrame.setLayout(new BorderLayout());
        toolFrame = menuFrame;
        floating = toolFrame == menuFrame;
        fIconkit = new Iconkit(toolFrame);
        if (icon != null) {
            Image iconImage = fIconkit.loadImageResource(icon);
            if (iconImage == null) {
               System.err.println("Resource icon "+icon+" could not be loaded!");
            } else {
               menuFrame.setIconImage(iconImage);
            }
        }
        Panel tools = createToolPalette();
        createTools(tools);
        toolFrame.add("Center", tools);

	MenuBar mb = new MenuBar();
	createMenus(mb);
	menuFrame.setMenuBar(mb);
        menuFrame.pack();

	// Workaround for Windows JDK pack()-bug:
	// The menu-frame is too small or too large.
        String windowResizable=System.getProperty("de.renew.windowResizable");
	if (windowResizable==null) {
            // If no parameter is given, the window is
            // resizable on Windows-Systems only.
	    if (System.getProperty("os.name").indexOf("Win")>=0) {
	       windowResizable="true";
            }
	}
        menuFrame.setResizable(Boolean.valueOf(windowResizable).booleanValue());

        //Dimension d = defaultSize();
	//setSize(d.width, d.height);
        menuFrame.show();
        if (toolFrame != menuFrame) {
           toolFrame.pack();
           toolFrame.show();
        } /* endif */

	windowCount++;
        addListeners();

        if (drawExtension==null)
            this.drawExtension=defaultDrawExtension;
        else    
           this.drawExtension=drawExtension;

	//System.out.println("number of params: "+drawings.length);
        Drawing drawing;
        for (int i=0; i<drawings.length; ++i)
           loadAndOpenDrawing(drawings[i]);
        if (fDrawings.size() == 0)
           promptNew();

        toolDone();

	if (dls!=null) {
	  dls.start();
	}
    }

    /**
     * Registers the listeners for the main windows
     */
    protected void addListeners() {
	    menuFrame.addWindowListener(
              new WindowAdapter() {
                public void windowClosing(WindowEvent event) {
                    exit();
                }
              }
            );
            if (toolFrame != menuFrame) {
	      toolFrame.addWindowListener(
                new WindowAdapter() {
                  public void windowClosing(WindowEvent event) {
		    toolFrame.setVisible(false);
                 }
                }
              );
            } /* endif */
    }

    public Frame menuFrame() { return menuFrame; }

    protected DrawApplication newWindow() { return new DrawApplication(menuFrame.getTitle()); }

//     private Menu titleSubstitute;

//     public synchronized void setTitle(String title) {
//       menuFrame.setTitle(title);
//       titleSubstitute.setLabel(title);
//     }

    private void updateName(Drawing drawing, String filename) {
      showDrawingViewFrame(drawing);
      fFilename = noExtension(filename);
      String nameOnly = noPath(fFilename);
      fViewFrame.setTitle(nameOnly);
      drawing.setName(nameOnly);
      drawing.setFilename(filename);
      int pos = fDrawings.indexOf(drawing);
      fDrawingsMenu.getItem(pos).setLabel(nameOnly);
    }

    /**
     * Creates the standard menus. Clients override this
     * method to add additional menus.
     */
    protected void createMenus(MenuBar mb) {
		mb.add(createFileMenu());
		mb.add(createEditMenu());
		mb.add(createAlignmentMenu());
		mb.add(createAttributesMenu());
		mb.add(createDebugMenu());
		mb.add(createDrawingsMenu());
    }

    void addDrawing(Drawing drawing) {
       if (!fDrawings.contains(drawing)) {
           fDrawings.addElement(drawing);
           CheckboxMenuItem mi = new CheckboxMenuItem(drawing.getName());
           mi.addItemListener(new DrawingMenuListener(this,drawing));
           fDrawingsMenu.add(mi);

	   drawingAdded(drawing);
       } /* endif */
    }

    private boolean closeViewFrame(DrawingViewFrame viewFrame) {
      Drawing drawing=viewFrame.view().drawing();

      if (drawing.isModified()) {
          int answer = MessageDialog.invokeDialog(
                viewFrame, "Renew: Confirm Close",
                new String[] {"The drawing \""+drawing.getName()+"\"","you are about to close","has been modified."},
                MessageDialog.LEFT, new String[] {" Save now "," Close ", " Cancel "}
          );
          if (answer == 0) {
             if (!saveDrawing(drawing))
                return false;
          } else if (answer == 2) {
            return false;
          }
      }
      viewFrame.setVisible(false);
      viewFrame.dispose();
      fViewFrames.removeElement(viewFrame);
      if (viewFrame == fViewFrame) {
        if (fViewFrames.size() > 0) {
	  setCurrentDrawing((DrawingViewFrame)fViewFrames.elementAt(fViewFrames.size()-1));
          fViewFrame.show();
fViewFrame.requestFocus(); // !!!
	} else
	  setCurrentDrawing(null);
      }

      int pos = fDrawings.indexOf(drawing);
      fDrawingsMenu.remove(pos);
      fDrawings.removeElementAt(pos);

      drawing.release();

      drawingReleased(drawing);

      viewFrame.discard();

      return true;
    }

    protected void drawingAdded(Drawing drawing) {
      // My subclasses will add special behaviour here.
    }

    protected void drawingReleased(Drawing drawing) {
      // My subclasses will add special behaviour here.
    }

    protected Menu createDrawingsMenu() {
        fDrawingsMenu = new Menu("Drawings");
        return fDrawingsMenu;
    }

    protected Enumeration drawings() {
      return fDrawings.elements();
    }

    private DrawingViewFrame newDrawingViewFrame(Drawing drawing) {
        Dimension d = getDrawingViewSize();
        StandardDrawingView drawingView = createDrawingView(d.width, d.height);
        addDrawing(drawing);
        DrawingViewFrame viewFrame=new DrawingViewFrame(this, drawingView, drawing);
        fViewFrames.addElement(viewFrame);
        setCurrentDrawing(viewFrame);
        return viewFrame;
    }
	  
    /**
     * Creates the file menu. Clients override this
     * method to add additional menu items.
     */
    protected Menu createFileMenu() {
		Menu menu = new Menu("File");
		MenuItem mi = new MenuItem("New Drawing", new MenuShortcut('n'));
		mi.addActionListener(
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
		            promptNew();
		        }
		    }
		);
		menu.add(mi);

// 		mi = new MenuItem("New Frame", new MenuShortcut('f'));
// 		mi.addActionListener(
// 		    new ActionListener() {
// 		        public void actionPerformed(ActionEvent event) {
// 			  newDrawingViewFrame(drawing());
// 		        }
// 		    }
// 		);
// 		menu.add(mi);

		mi = new MenuItem("Open Drawing...", new MenuShortcut('o'));
		mi.addActionListener(
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
		            promptOpen();
		        }
		    }
		);
		menu.add(mi);

		mi = new MenuItem("Save Drawing", new MenuShortcut('s'));
		mi.addActionListener(
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
		            saveDrawing(drawing());
		        }
		    }
		);
		menu.add(mi);

		mi = new MenuItem("Save Drawing As...");
		mi.addActionListener(
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
                            Drawing drawing=drawing();
		            if (promptSaveAs(drawing))
                                saveDrawing(drawing);
		        }
		    }
		);
		menu.add(mi);

		mi = new MenuItem("Save All Drawings");
		mi.addActionListener(
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
                            Enumeration drawings=drawings();
                            while (drawings.hasMoreElements()) {
                               Drawing drawing=(Drawing)drawings.nextElement();
                               if (drawing.getFilename()==null) {
                                  showDrawingViewFrame(drawing);
                                  if (!promptSaveAs(drawing))
                                      return;
                               }
                             }
                             drawings=drawings();
                             while (drawings.hasMoreElements()) {
		                saveDrawing((Drawing)drawings.nextElement());
                            }
		        }
		    }
		);
		menu.add(mi);

    	mi = new MenuItem("Close Drawing");
       	mi.addActionListener(
	   new ActionListener() {
	      public void actionPerformed(ActionEvent event) {
                 if (fViewFrame != null) {
         	    closeViewFrame(fViewFrame);
         	 }
	      }
	   }
	);
	menu.add(mi);

//     	mi = new MenuItem("Close Frame");
//        	mi.addActionListener(
// 	   new ActionListener() {
// 	      public void actionPerformed(ActionEvent event) {
// 		fViewFrame.hide();
//                 fViewFrame.dispose();
// 	      }
// 	   }
// 	);
// 	menu.add(mi);

		menu.addSeparator();
		mi = new MenuItem("Print...", new MenuShortcut('p'));
		mi.addActionListener(
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
		            print();
		        }
		    }
		);
		menu.add(mi);
		menu.addSeparator();
		mi = new MenuItem("Exit");
		mi.addActionListener(
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
		            exit();
		        }
		    }
		);
		menu.add(mi);
		return menu;
	}

    /**
     * Creates the edit menu. Clients override this
     * method to add additional menu items.
     */
    protected Menu createEditMenu() {
		CommandMenu menu = new CommandMenu("Edit");
		menu.add(new CutCommand("Cut", this), new MenuShortcut('x'));
		menu.add(new CopyCommand("Copy", this), new MenuShortcut('c'));
		menu.add(new PasteCommand("Paste", this), new MenuShortcut('v'));
		menu.addSeparator();
		menu.add(new DuplicateCommand("Duplicate", this), new MenuShortcut('d'));
		menu.add(new DeleteCommand("Delete", this));
		menu.addSeparator();
		menu.add(new GroupCommand("Group", this));
		menu.add(new UngroupCommand("Ungroup", this));
		menu.addSeparator();
		menu.add(new SendToBackCommand("Send to Back", this));
		menu.add(new BringToFrontCommand("Bring to Front", this));
		menu.addSeparator();
                menu.add(new SelectAllCommand("Select All", this), new MenuShortcut('a'));
		menu.addSeparator();
		menu.add(new ToggleStickyCommand("Toggle Sticky Tools",
		  this));
		return menu;
	}

    /**
     * Creates the alignment menu. Clients override this
     * method to add additional menu items.
     */
    protected Menu createAlignmentMenu() {
		CommandMenu menu = new CommandMenu("Align");
                menu.add(new ToggleGridCommand("Toggle Snap to Grid",
		  this, new Point(5,5)));
		menu.addSeparator();
		menu.add(new AlignCommand("Lefts", this, AlignCommand.LEFTS));
		menu.add(new AlignCommand("Centers", this, AlignCommand.CENTERS));
		menu.add(new AlignCommand("Rights", this, AlignCommand.RIGHTS));
		menu.addSeparator();
		menu.add(new AlignCommand("Tops", this, AlignCommand.TOPS));
		menu.add(new AlignCommand("Middles", this, AlignCommand.MIDDLES));
		menu.add(new AlignCommand("Bottoms", this, AlignCommand.BOTTOMS));
		return menu;
	}

    /**
     * Creates the debug menu. Clients override this
     * method to add additional menu items.
     */
    protected Menu createDebugMenu() {
		Menu menu = new Menu("Debug");

		MenuItem mi = new MenuItem("Simple Update");
		mi.addActionListener(
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
		            view().setDisplayUpdate(new SimpleUpdateStrategy());
		        }
		    }
		);
		menu.add(mi);

		mi = new MenuItem("Buffered Update");
		mi.addActionListener(
		    new ActionListener() {
		        public void actionPerformed(ActionEvent event) {
		            view().setDisplayUpdate(new BufferedUpdateStrategy());
		        }
		    }
		);
		menu.add(mi);
		return menu;
	}

    /**
     * Creates the attributes menu and its submenus. Clients override this
     * method to add additional menu items.
     */
    protected Menu createAttributesMenu() {
        Menu menu = new Menu("Attributes");
        menu.add(createColorMenu("Fill Color", "FillColor"));
        menu.add(createColorMenu("Pen Color", "FrameColor"));
        menu.add(createArrowMenu());
		menu.addSeparator();
        menu.add(createFontMenu());
        menu.add(createFontSizeMenu());
        menu.add(createFontStyleMenu());
        menu.add(createColorMenu("Text Color", "TextColor"));
        return menu;
    }

    /**
     * Creates the color menu.
     */
    protected Menu createColorMenu(String title, String attribute) {
        CommandMenu menu = new CommandMenu(title);
        for (int i=0; i<ColorMap.size(); i++)
            menu.add(
                new ChangeAttributeCommand(
                    ColorMap.name(i),
                    attribute,
                    ColorMap.color(i),
                    this
                )
            );
        return menu;
    }

    /**
     * Creates the arrows menu.
     */
    protected Menu createArrowMenu() {
        CommandMenu menu = new CommandMenu("Arrow");
        menu.add(new ChangeAttributeCommand("none",     "ArrowMode", new Integer(PolyLineFigure.ARROW_TIP_NONE),  this));
        menu.add(new ChangeAttributeCommand("at Start", "ArrowMode", new Integer(PolyLineFigure.ARROW_TIP_START), this));
        menu.add(new ChangeAttributeCommand("at End",   "ArrowMode", new Integer(PolyLineFigure.ARROW_TIP_END),   this));
        menu.add(new ChangeAttributeCommand("at Both",  "ArrowMode", new Integer(PolyLineFigure.ARROW_TIP_BOTH),  this));
        return menu;
    }

    /**
     * Creates the fonts menus. It installs all available fonts
     * supported by the toolkit implementation.
     */
    protected Menu createFontMenu() {
        CommandMenu menu = new CommandMenu("Font");
        String fonts[] = Toolkit.getDefaultToolkit().getFontList();
        for (int i = 0; i < fonts.length; i++)
            menu.add(new ChangeAttributeCommand(fonts[i], "FontName", fonts[i],  this));
        return menu;
    }

    /**
     * Creates the font style menu with entries (Plain, Italic, Bold).
     */
    protected Menu createFontStyleMenu() {
        CommandMenu menu = new CommandMenu("Font Style");
        menu.add(new ChangeAttributeCommand("Plain", "FontStyle", new Integer(Font.PLAIN), this));
        menu.add(new ChangeAttributeCommand("Italic","FontStyle", new Integer(Font.ITALIC),this));
        menu.add(new ChangeAttributeCommand("Bold",  "FontStyle", new Integer(Font.BOLD),  this));
        return menu;
    }

    /**
     * Creates the font size menu.
     */
    protected Menu createFontSizeMenu() {
        CommandMenu menu = new CommandMenu("Font Size");
        int sizes[] = { 9, 10, 12, 14, 18, 24, 36, 48, 72 };
        for (int i = 0; i < sizes.length; i++) {
            menu.add(
                new ChangeAttributeCommand(
                    Integer.toString(sizes[i]),
                    "FontSize",
                    new Integer(sizes[i]),  this)
                );
        }
        return menu;
    }

    /**
     * Creates the tool palette.
     */
    protected Panel createToolPalette() {
        Panel palette = new Panel();
        palette.setBackground(Color.lightGray);
        palette.setLayout(new PaletteLayout(2,new Point(2,2),false));
        return palette;
    }

    /**
     * Creates the tools. By default all standard figure creation tools are added.
     * Override this method to remove the standard tools.
     * Call the inherited method to include the standard tools.
     * @param palette the palette where the tools are added.
     */
    protected void createTools(Panel palette) {
        Tool tool = createSelectionTool();

        fDefaultToolButton = createToolButton(IMAGES+"SEL", "Selection Tool", tool);
        palette.add(fDefaultToolButton);

        createAdditionalTools(palette);

        tool = new CreationTool(this, new RectangleFigure());
        palette.add(createToolButton(IMAGES+"RECT", "Rectangle Tool", tool));

        tool = new CreationTool(this, new RoundRectangleFigure());
        palette.add(createToolButton(IMAGES+"RRECT", "Round Rectangle Tool", tool));

        tool = new CreationTool(this, new EllipseFigure());
        palette.add(createToolButton(IMAGES+"ELLIPSE", "Ellipse Tool", tool));

        tool = new CreationTool(this, new DiamondFigure());
        palette.add(createToolButton(IMAGES+"DIAMOND", "Diamond Tool", tool));

        tool = new CreationTool(this, new TriangleFigure());
        palette.add(createToolButton(IMAGES+"TRIANGLE", "Triangle Tool", tool));

        tool = new CreationTool(this, new LineFigure());
        palette.add(createToolButton(IMAGES+"LINE", "Line Tool", tool));

        tool = new ConnectionTool(this, new LineConnection());
        palette.add(createToolButton(IMAGES+"CONN", "Connection Tool", tool));

        tool = new ConnectionTool(this, new ElbowConnection());
        palette.add(createToolButton(IMAGES+"OCONN", "Elbow Connection Tool", tool));

        tool = new ScribbleTool(this);
        palette.add(createToolButton(IMAGES+"SCRIBBL", "Scribble Tool", tool));

        tool = new PolygonTool(this);
        palette.add(createToolButton(IMAGES+"POLYGON", "Polygon Tool", tool));

        tool = new ImageFigureCreationTool(this);
        palette.add(createToolButton(IMAGES+"IMAGE", "Image Tool", tool));

        tool = new TextTool(this, new TextFigure());
        fTextTB = createToolButton(IMAGES+"TEXT", "Text Tool", tool);
        palette.add(fTextTB);

        tool = new ConnectedTextTool(this, new TextFigure());
        fConnTextTB = createToolButton(IMAGES+"ATEXT","Connected Text Tool",tool);
        palette.add(fConnTextTB);

    }


    /**
     * Override this method to add additional tools.
     */
    protected void createAdditionalTools(Panel palette) { }


    /**
     * Creates the selection tool used in this editor. Override to use
     * a custom selection tool.
     */
    protected Tool createSelectionTool() {
        return new SelectionTool(this);
    }

    /**
     * Creates a tool button with the given image, tool, and text
     */
    protected ToolButton createToolButton(String iconName, String toolName, Tool tool) {
        return new ToolButton(this, iconName, toolName, tool);
    }

    /**
     * Creates the drawing view used in this application.
     * You need to override this method to use a DrawingView
     * subclass in your application. By default a standard
     * DrawingView is returned.
     */
    protected StandardDrawingView createDrawingView(int width, int height) {
        return new StandardDrawingView(this, width, height);
    }

    void unsetCurrentDrawing() { 
      if (fViewFrame!=null) {
	int pos = fDrawings.indexOf(drawing());
	if (pos >= 0) {
	  ((CheckboxMenuItem)fDrawingsMenu.getItem(pos)).setState(false);
	}
	fViewFrame=null;
      }
    }

    void setCurrentDrawing(DrawingViewFrame viewFrame) {
      if (viewFrame==null) {
	unsetCurrentDrawing();
      } else {
	Drawing frameDrawing=viewFrame.view().drawing();
//System.out.println("Trying to activate drawing "+frameDrawing.getName());
	int pos=fDrawings.indexOf(frameDrawing);
	
	// Is the drawing still active or do we process a delayed event?
	if (pos>=0 && fViewFrames.removeElement(viewFrame)) {
          fViewFrames.addElement(viewFrame);
	  unsetCurrentDrawing();
          // new: move viewFrame to last position in vector
          // end new
	  fViewFrame = viewFrame;
	  ((CheckboxMenuItem)fDrawingsMenu.getItem(pos)).setState(true);
//System.out.println("Activated drawing "+frameDrawing.getName());
//displayZOrder();
	}
      }
    }

/* *****
private void displayZOrder() {
   for (int i=0; i<fViewFrames.size(); ++i) {
      System.out.println(i+". "+((DrawingViewFrame)fViewFrames.elementAt(i)).view().drawing().getName());
   }
}
***** */

    private DrawingViewFrame getViewFrame(Drawing drawing) {
      Enumeration viewFrames = fViewFrames.elements();
      while (viewFrames.hasMoreElements()) {
	  DrawingViewFrame viewFrame =
            (DrawingViewFrame)viewFrames.nextElement();
	  if (viewFrame.view().drawing() == drawing)
                return viewFrame;
       }
       return null;
    }

    public DrawingViewFrame showDrawingViewFrame(Drawing drawing) {
        DrawingViewFrame viewFrame = getViewFrame(drawing);
        if (viewFrame != null) {
            setCurrentDrawing(viewFrame);
            viewFrame.show();
            viewFrame.requestFocus(); // was missing!!!
        } else {
          // Nothing was found. Create a new frame.
          viewFrame = newDrawingViewFrame(drawing);
        }
        return viewFrame;
    }

    public boolean closeDrawing(Drawing drawing) {
        DrawingViewFrame viewFrame = getViewFrame(drawing);
        if (viewFrame != null) {
           return closeViewFrame(viewFrame);
        }
        return false;
    }

    /**
     * Override to define the dimensions of the drawing view.
     */
    protected Dimension getDrawingViewSize() {
        return new Dimension(400, 600);
    }

    /**
     * Creates the drawing used in this application.
     * You need to override this method to use a Drawing
     * subclass in your application. By default a standard
     * Drawing is returned.
     */
    protected Drawing createDrawing() {
        return new StandardDrawing();
    }

    /**
     * Creates the status line.
     */
    protected TextField createStatusLine() {
        TextField field = new TextField("No Tool", 40);
        field.setEditable(false);
        return field;
    }

    /**
     * Handles a user selection in the palette.
     * @see PaletteListener
     */
    public void paletteUserSelected(PaletteButton button) {
        ToolButton toolButton = (ToolButton) button;
        setSelected(toolButton);
        setTool(toolButton.tool(), toolButton.name());
    }

    /**
     * Handles when the mouse enters or leaves a palette button.
     * @see PaletteListener
     */
    public void paletteUserOver(PaletteButton button, boolean inside) {
        ToolButton toolButton = (ToolButton) button;
        if (inside)
            showStatus(toolButton.name());
        else if (fSelectedToolButton != null)
            showStatus(fSelectedToolButton.name());
    }

    /**
     * Gets the current drawing.
     * @see DrawingEditor
     */
    public Drawing drawing() {
        return view().drawing();
    }

    /**
     * Gets the current tool.
     * @see DrawingEditor
     */
    public Tool tool() {
        return fTool;
    }

    /**
     * Gets the current drawing view.
     * @see DrawingEditor
     */
    public DrawingView view() {
      DrawingViewFrame viewFrame=fViewFrame;
      if (viewFrame==null) {
	return NullDrawingView.INSTANCE;
      } else {
	return viewFrame.view();
      }
    }

    /**
     * Toggle the sticky mode.
     */
    public void toggleToolButtonSticky() {
      fToolButtonSticky=!fToolButtonSticky;
    }

    /**
     * Sets the default tool of the editor.
     * @see DrawingEditor
     */
    public void toolDone() {
      // not while initializing!
      if (windowCount > 0) {
	if (fToolButtonSticky && tool()!=null) {
	  tool().deactivate();
	  tool().activate();
	} else if (fDefaultToolButton!=null) {
	  setTool(fDefaultToolButton.tool(), fDefaultToolButton.name());
	  setSelected(fDefaultToolButton);
        }
      }
    }

    /**
     * Handles a change of the current selection. Updates all
     * menu items that are selection sensitive.
     * @see DrawingEditor
     */
    public void selectionChanged(DrawingView view) {
        MenuBar mb = menuFrame.getMenuBar();
        CommandMenu editMenu = (CommandMenu)mb.getMenu(EDIT_MENU);
        editMenu.checkEnabled();
        CommandMenu alignmentMenu = (CommandMenu)mb.getMenu(ALIGNMENT_MENU);
        alignmentMenu.checkEnabled();
    }

    /**
     * Shows a status message.
     * @see DrawingEditor
     */
    public void showStatus(String string) {
        fStatusLine.setText(string);
    }

    private void setTool(Tool t, String name) {
        if (fTool != null)
            fTool.deactivate();
        fTool = t;
        if (fTool != null) {
            fStatusLine.setText(name);
            fTool.activate();
        }
    }

    private void setSelected(ToolButton button) {
        if (fSelectedToolButton != null)
            fSelectedToolButton.reset();
        fSelectedToolButton = button;
        if (fSelectedToolButton != null)
            fSelectedToolButton.select();
    }


    public ToolButton toolButtonForTextFigure(TextFigure figure) {
       if (figure.parent() == null) {
          return fTextTB;
       } else {
          return fConnTextTB;
       }
    }

    public void doTextEdit(TextFigure textFigure, int line, int column) {
       ToolButton tb=toolButtonForTextFigure(textFigure);
       paletteUserSelected(tb);
       TextTool tt = (TextTool)tb.tool();
       tt.beginEdit(textFigure);
       if (line > 0 && column > 0) {
	    tt.setCaretPosition(line,column);
       }
    }

    public void doTextEdit(TextFigure textFigure) {
       doTextEdit(textFigure, 0,0);
    }

    /**
     * Exits the application. You should never override this method
     */
    public void exit() {
        // Close all DrawingViewFrames:
        while (true) {
            Enumeration viewFrames = fViewFrames.elements();
            if (!viewFrames.hasMoreElements()) {
               break;
            } else {
              DrawingViewFrame viewFrame =
                 (DrawingViewFrame)viewFrames.nextElement();
              if (!closeViewFrame(viewFrame))
                  return;
            }
        }
        destroy();
        toolFrame.setVisible(false);
        toolFrame.dispose();
        menuFrame.setVisible(false);      // hide the Frame
        menuFrame.dispose();   // tell windowing system to free resources
        if (--windowCount <= 0)
           System.exit(0);
    }

    /**
     * Handles additional clean up operations. Override to destroy
     * or release drawing editor resources.
     */
    protected void destroy() {
    }



    /**
     * Creates a new drawing view with a new empty drawing.
     */
    public void promptNew() {
       Drawing drawing = createDrawing();
       newDrawingViewFrame(drawing);
    }

    /**
     * Shows a file dialog and opens a drawing.
     */
    public void promptOpen() {
        FileDialog dialog = new FileDialog(menuFrame, "Open File...", FileDialog.LOAD);
        dialog.show();
        String filename = dialog.getFile();
        if (filename != null) {
            filename = stripTrailingAsterisks(filename);
            String dirname = dialog.getDirectory();
            Drawing drawing;
            loadAndOpenDrawing(dirname + filename);
        }
        dialog.dispose();
    }

    public static String noExtension(String path) {
      int end = path.lastIndexOf(".");
      if (end >= 0)
          return path.substring(0,end);
      else
          return path;
    }

    public static String noPath(String path) {
      int start = path.lastIndexOf(System.getProperty("file.separator"));
      // if (start == -1) start = 0;
      return path.substring(start+1);
    }

    public static String pathOnly(String path) {
      int end = path.lastIndexOf(System.getProperty("file.separator"));
      if (end == -1)
          return "";
      return path.substring(0,end+1);
    }

   public boolean saveDrawing(Drawing drawing) {
      String filename=drawing.getFilename();
      if (filename==null) {
         if (!promptSaveAs(drawing))
            return false;
         filename=drawing.getFilename();
      }
      saveDrawing(drawing,filename);
      return true;
   }

   public void saveDrawing(Drawing drawing, String filename) {
         String msg="Saving drawing "+drawing.getName()+" as "+filename+"...";
         System.out.println(msg);
         showStatus(msg);
         if (filename.endsWith(drawExtension)) {
            saveAsStorableOutput(drawing,filename);
         } else {
            saveAsObjectOutput(drawing,filename);
         }
    }

    /**
     * Shows a file dialog for saving a drawing.
     * Returns null or a complete file name.
     * If a file name is determined, the drawing's name is set accordingly.
     */
    boolean promptSaveAs(Drawing drawing) {
        String filename=drawing.getFilename();
        if (filename==null)
           filename=pathOnly(fFilename)+drawing.getName()+drawExtension;
        toolDone();
        filename = getSavePath("Save File...",filename);
        if (filename == null)
           return false;
        if (!filename.endsWith(drawExtension) && !filename.endsWith(".ser"))
            filename += drawExtension;
        updateName(drawing,filename);
        return true;
    }

    /**
     * Prints the drawing.
     */
    public void print() {
        fTool.deactivate();
        PrintJob printJob = menuFrame.getToolkit().getPrintJob(fViewFrame, "Print Drawing", null);

        if (printJob != null) {
            Graphics pg = printJob.getGraphics();

            if (pg != null) {
	        // OK: This was 
	        // fViewFrame.drawingView.printAll(pg);
	        // which relied on a standard drawing view.
	        // There is no reason why a call to the standard
	        // paint method should not succeed.
	        view().paint(pg);
                pg.dispose(); // flush page
            }
            printJob.end();
        }
        fTool.activate();
    }

    protected String getSavePath(String title, String defaultName) {
        String path = null;
        FileDialog dialog = new FileDialog(menuFrame, title, FileDialog.SAVE);
	if (defaultName != null)
	  dialog.setFile(defaultName);
        dialog.show();
        String filename = dialog.getFile();
        if (filename != null) {
            filename = stripTrailingAsterisks(filename);
            String dirname = dialog.getDirectory();
            path = dirname + filename;
        }
        dialog.dispose();
        return path;
    }

    private String stripTrailingAsterisks(String filename) {
        // workaround for bug on NT
        if (filename.endsWith("*.*"))
            return filename.substring(0, filename.length() - 4);
        else
            return filename;
    }

    public static final int FILEVERSION=1;

    private void saveAsStorableOutput(Drawing drawing, String file) {
        // TBD: should write a MIME header
        try {
            FileOutputStream stream = new FileOutputStream(file);
            StorableOutput output = new StorableOutput(stream);
	    // Write the current version. This is a
	    // substitute for a MIME header.
	    // Changes version 0->1:
	    //  - PolyLineFigure is now an AttributeFigure.
	    output.writeInt(FILEVERSION);
	    // Write the drawing.
            output.writeStorable(drawing);
            output.close();
        } catch (IOException e) {
            showStatus(e.toString());
        }
    }

    private void saveAsObjectOutput(Drawing drawing, String file) {
        // TBD: should write a MIME header
        try {
            FileOutputStream stream = new FileOutputStream(file);
            ObjectOutput output = new ObjectOutputStream(stream);
            output.writeObject(drawing);
            output.close();
        } catch (IOException e) {
            showStatus(e.toString());
        }
    }

    public synchronized void loadAndOpenDrawing(String file) {
       Drawing drawing  = loadDrawing(file);
       if (drawing != null)
	     newDrawingViewFrame(drawing);
    }


    private Drawing loadDrawing(String file) {
        String msg="Loading drawing from "+file+"...";
        System.out.println(msg);
        showStatus(msg);
        Drawing drawing = null;
//        toolDone();
        String type = guessType(file);
        if (type.equals("storable"))
            drawing = readFromStorableInput(file);
        else if (type.equals("serialized"))
            drawing = readFromObjectInput(file);
        else {
            showStatus("Unknown file type of file \""+file+"\"");
            System.out.println("Unknown file type of file \""+file+"\"");
        }
        if (drawing != null) {
	   drawing.setName(noPath(noExtension(file)));
           drawing.setFilename(file);
           fFilename = noExtension(file);
           if (drawing instanceof Figure) {
              ((Figure)drawing).invalidate();
           }
        }
        return drawing;
    }

    /**
     * Override to establish a new policy to create
     * input handlers.
     */

    protected StorableInput makeStorableInput(FileInputStream stream,
      boolean useUFT)
    {
      return new StorableInput(stream,useUFT);
    }

    private Drawing readFromStorableInput(String file) {
        Drawing drawing = null;
        try {
            FileInputStream stream = new FileInputStream(file);
            StorableInput input = makeStorableInput(stream,true);
            try {
              drawing = readStorableDrawing(input);
            } catch (sun.io.MalformedInputException e) {
               stream.close();
               stream = new FileInputStream(file);
               input = makeStorableInput(stream,false);
               drawing = readStorableDrawing(input);
            }
        } catch (IOException e) {
            showStatus("Error: " + e);
            System.out.println("Error: " + e);
        }
        return drawing;
    }

    private Drawing readStorableDrawing(StorableInput input)
      throws IOException
    {
      if (input.canReadInt()) {
	input.setVersion(input.readInt());
      } else {
	input.setVersion(0);
      }
      
      return (Drawing)input.readStorable();
    }

    private Drawing readFromObjectInput(String file) {
        Drawing drawing = null;
        try {
            FileInputStream stream = new FileInputStream(file);
            ObjectInput input = new ObjectInputStream(stream);
            drawing = (Drawing)input.readObject();
        } catch (IOException e) {
            showStatus("Error: " + e);
            System.out.println("Error: " + e);
        } catch (ClassNotFoundException e) {
            showStatus("Class not found: " + e);
            System.out.println("Error: " + e);
        }
        return drawing;
    }

    private String guessType(String file) {
        if (file.endsWith(drawExtension))
            return "storable";
        if (file.endsWith(".ser"))
            return "serialized";
        return "unknown";
    }

    public void windowOpened(WindowEvent e) {
       //DrawingViewFrame viewFrame = (DrawingViewFrame)e.getWindow();
       //fViewFrames.addElement(viewFrame);
       //System.out.println("New Drawing View Window: "+viewFrame.getTitle());
    }

    public void windowActivated(WindowEvent e) {
      DrawingViewFrame viewFrame = (DrawingViewFrame)e.getWindow();
      if (viewFrame != null && viewFrame != fViewFrame) {
	//System.out.println("Current Drawing View Window: "+viewFrame.getTitle());
      }
      setCurrentDrawing(viewFrame);
    }

    public void windowDeactivated(WindowEvent e) {    }

    public void windowClosing(WindowEvent e) {
      DrawingViewFrame viewFrame = (DrawingViewFrame)e.getWindow();
      // !!!!!!!!!!! This was:
      // removeDrawing(viewFrame.drawingview.drawing());
      closeViewFrame(viewFrame);
    }

    public void windowClosed(WindowEvent e) {
    }

    public void windowIconified(WindowEvent e) {
    }

    public void windowDeiconified(WindowEvent e) {
    }

    //-- main -----------------------------------------------------------

    public static void main(String[] args) {
	new DrawApplication("JHotDraw",args);
    }

}

class DrawingMenuListener implements ItemListener {
   private DrawApplication fEditor;
   private Drawing fDrawing;

   DrawingMenuListener(DrawApplication editor, Drawing drawing) {
      fEditor = editor;
      fDrawing = drawing;
   }

   public void itemStateChanged(ItemEvent event) {
      fEditor.showDrawingViewFrame(fDrawing);
   }
}

class DrawingLoadServer extends Thread {

        private DrawApplication editor;
        private ServerSocket s=null;

        public DrawingLoadServer(DrawApplication editor, int port) throws IOException {
            this.editor=editor;
            System.out.println("Drawing Load Server setting up server socket at port "+port+"...");
            s=new ServerSocket(port);
        }

        public void run() {
            System.out.println("Drawing Load Server waiting for parameters...");
            while (true) {
              Socket client=null;
              try {
                 client=s.accept();
                 System.out.println("Parameter server accepted client.");
                 BufferedReader in=new BufferedReader(new InputStreamReader(client.getInputStream()));
                 try {
                    String drawingFileName;
                    do {
                         drawingFileName=in.readLine();
                         if (drawingFileName!=null) {
                            System.out.println("Received Parameter "+drawingFileName);
                            editor.loadAndOpenDrawing(drawingFileName);
                         }
                    } while (drawingFileName != null);
                 } catch (IOException e) { }
                 System.out.println("Connection closed.");
                 in.close();
                 client.close();
              } catch (Exception e) {
                 System.out.println("Drawing Load Server threw exception: "+e);
              }
           }
        }
}
