package de.renew.function;

import java.lang.reflect.*;
import de.renew.util.Types;

public class Executor {
  /** returned if there is no result from a method call (void method) */
  public final static Object VOIDRETURN=new Object(); 

  /** This class cannot to be instantiated. */
  private Executor() {}

  /**
   * Get the types associated to an array of objects.
   */  
  static public Class[] getTypes(Object[] params) 
  {
    Class[] parameterTypes=new Class[params.length];
    for (int i=0; i<params.length; i++) {
      if (params[i] instanceof Value) {
	parameterTypes[i]=Types.typify(
          ((Value)params[i]).value.getClass());
      } else {
	parameterTypes[i]=params[i].getClass();
      }
    }
    return parameterTypes;
  }

  /**
   * Remove the valueness of the argument and convert it
   * to the desired class, if that is possible without
   * loss of information.
   */
  static public Object unvalueAndCast(Object obj, Class clazz) 
    throws IllegalArgumentException
  {
    if (obj instanceof Value) {
      if (clazz.isPrimitive()) {
	Object value=((Value)obj).value;

	// Try to return characters, if possible.
	if (value instanceof Character) {
	  if (clazz==Character.TYPE) {
	    return value;
	  } else {
	    // No, cast the character to an integer and
	    // process it later.
	    value=new Integer(((Character)value).charValue());
	  }
	}
	
	if (value instanceof Number) {
	  Number number=(Number)value;
	  if (clazz==Double.TYPE) {
	    return new Double(number.doubleValue());
	  }
	  if (number instanceof Double) {
	    throw new IllegalArgumentException();
	  }
	  if (clazz==Float.TYPE) {
	    return new Float(number.floatValue());
	  }
	  if (number instanceof Float) {
	    throw new IllegalArgumentException();
	  }
	  if (clazz==Long.TYPE) {
	    return new Long(number.longValue());
	  }
	  if (number instanceof Long) {
	    throw new IllegalArgumentException();
	  }
	  if (clazz==Integer.TYPE) {
	    return new Integer(number.intValue());
	  }
	  if (number instanceof Integer) {
	    throw new IllegalArgumentException();
	  }
	  if (clazz==Short.TYPE) {
	    return new Short(number.shortValue());
	  }
	  if (number instanceof Short) {
	    throw new IllegalArgumentException();
	  }
	  if (clazz==Byte.TYPE) {
	    return new Byte(number.byteValue());
	  }
	  throw new IllegalArgumentException();
	} else if (value instanceof Boolean) {
	  if (clazz==Boolean.TYPE) {
	    return value;
	  } else {
	    throw new IllegalArgumentException();
	  }	  
	} else {
	  throw new RuntimeException("Encountered a bad value.");
	}
      } else {
	throw new IllegalArgumentException();
      }
    } else {
      if (obj==null || clazz.isInstance(obj)) {
	return obj;
      } else {
	throw new IllegalArgumentException();
      }
    }
  }

  /**
   * Wrap an object in a value, if neccessary.
   */
  static public Object possiblyWrap(Object obj, boolean wrap) {
    if (wrap) {
      obj=new Value(obj);
    }
    return obj;
  }

  /**
   * Unvalue and cast a complete array.
   */
  static public Object[] unvalueAndCast(Object[] objs, Class[] clazzes)
    throws IllegalArgumentException
  {
    Object[] results=new Object[objs.length];
    for (int i=0;i<objs.length;i++) {
      results[i]=unvalueAndCast(objs[i],clazzes[i]);
    }
    return results;
  }

  /**
   * Given an array of actual parameter types, find one of 
   * the most exact match in an array of constructors. If multiple
   * incomparable matches are found, the first match
   * is returned.
   */
  static public Constructor findBestConstructor(Class clazz, 
    Class[] params) throws NoSuchMethodException
  {
    // Find the constructor with exact types.
    try {
      return clazz.getConstructor(getTypes(params));
    }
    catch (Exception e){
      // This is expected: the constructor does not exist.
      // Let's try it the hard way.
    }

    // Make a list of all constructors and search for the
    // optimal one.
    Constructor[] constructors=clazz.getConstructors();

    Constructor best=null;
    Class[] bestParams=null;

    for (int i=0;i<constructors.length;i++) {
      Class[] currentParams=constructors[i].getParameterTypes();
      if (Types.allowsWideningConversion(params,currentParams)) {
	if (best==null ||
	  Types.allowsWideningConversion(currentParams,bestParams))
	{
	  best=constructors[i];
	  bestParams=currentParams;
	}
      }
    }

    if (best==null) {
      throw new NoSuchMethodException();
    }
    
    return best;
  }

  /**
   * Given an array of actual parameter types, find one of 
   * the most exact match in an array of methods. If multiple
   * incomparable matches are found, the first match
   * is returned.
   */
  static public Method findBestMethod(Class clazz, 
    String name, Class[] params) throws NoSuchMethodException
  {
    // Find the method with exact types.
    try {
      return clazz.getMethod(name,getTypes(params));
    }
    catch (Exception e){
      // This is expected: the method does not exist.
      // Let's try it the hard way.
    }

    Method[] methods=clazz.getMethods();
    if (clazz.isInterface()) {
      // We must also look at the methods of java.lang.Object.
      // This is special for methods, as constructors are never called
      // for an interface.
      Method[] interfaceMethods=methods;
      Method[] objectMethods=Object.class.getMethods();
      methods=new Method[interfaceMethods.length+objectMethods.length];
      System.arraycopy(interfaceMethods,0,
        methods,0,interfaceMethods.length);
      System.arraycopy(objectMethods,0,
        methods,interfaceMethods.length,objectMethods.length);
    }

    Method best=null;
    Class[] bestParams=null;

    for (int i=0;i<methods.length;i++) {
      if (name.equals(methods[i].getName())) {
	Class[] currentParams=methods[i].getParameterTypes();
	if (Types.allowsWideningConversion(params,currentParams)) {
	  if (best==null ||
	    Types.allowsWideningConversion(currentParams,bestParams))
	  {
	    best=methods[i];
	    bestParams=currentParams;
	  }
	}
      }
    }

    if (best==null) {
      throw new NoSuchMethodException();
    }
    
    return best;
  }

  /**
   * Execute a constructor with parameter objects.
   */
  static public Object executeConstructor(Constructor constructor,
    Object[] params) throws NoSuchMethodException, InvocationTargetException
  {
    Object result=null;
    try {
      Object[] castedParams=unvalueAndCast(params,
	constructor.getParameterTypes());
      result=constructor.newInstance(castedParams);
    }
    catch (IllegalArgumentException ex) {
      // This should never happen as we got the right number and
      // types of parameters because we got this constructor object
      // through reflections.
      throw new RuntimeException("IllegalArgumentException in Executor!");
    }
    catch (InstantiationException ex) {
      // This is thrown if the class that is about to be
      // instantiated is abstract. In this case, 
      // we throw an NoSuchMethodException. It is in fact a
      // constructor that cannot be called directly.
      throw new NoSuchMethodException(ex.toString());
    }
    catch (IllegalAccessException ex) {
      // This is thrown if the constructor that should be called
      // is not public. As only public method should be known in
      // the rest of the world, we throw a NoSuchMethodException.
      throw new NoSuchMethodException(ex.toString());
    }
    return result;
  }

  /**
   * Execute a constructor given the target class and the
   * parameter objects.
   */
  static public Object executeConstructor(Class clazz, Object[] params)
    throws NoSuchMethodException, InvocationTargetException
  {
    if (clazz.isArray()) {
      int[] dimensions=new int[params.length];
      try {
        for (int i=0;i<params.length;i++) {
	  dimensions[i]=
            ((Integer)unvalueAndCast(params[i],Integer.TYPE)).intValue();
	  clazz=clazz.getComponentType();
        }
	if (clazz==null) {
	  throw new Exception();
	}
      }
      catch (Exception e) {
	throw new NoSuchMethodException("Bad array constructor.");
      }
      
      return Array.newInstance(clazz,dimensions);
    } else {
      Constructor constructor=findBestConstructor(clazz,getTypes(params));
      return executeConstructor(constructor,params);
    }
  }

  /**
   * Execute a given method on an object using some parameters.
   *
   * The target must be null if and only if a static method is to be called.
   */
  static public Object executeMethod(Method method, Object target,
    Object[] params) throws NoSuchMethodException,
    InvocationTargetException
  {
    Object result=null;
    try {
      Object[] castedParams=unvalueAndCast(params,
	method.getParameterTypes());
      result=method.invoke(target,castedParams);
      if (method.getReturnType()==Void.TYPE) {
	result=VOIDRETURN;
      } else {
	result=possiblyWrap(result,method.getReturnType().isPrimitive());
      }
    }
    catch (IllegalArgumentException ex) {
      // this should never happen as we got the right number and types of
      // parameters because we got this method object through reflections
      throw new RuntimeException(
        "IllegalArgumentException in Executor!");
    }
    catch (IllegalAccessException ex) {
      // This is thrown if the method that should be called is not public.
      // As only public method should be known in the rest of the world,
      // we throw a NoSuchMethodException
      throw new NoSuchMethodException(ex.toString());
    }
 
    return result;
  }

  /**
   * Execute a method of a given name in a given class on an object
   * using some parameters.
   *
   * The target object should be of the provided clazz.
   * It must be null if and only if a static method is to be called.
   */
  static public Object executeMethod(Class clazz, Object target,
    String name, Object[] params) throws NoSuchMethodException,
    InvocationTargetException
  {
    Method method=findBestMethod(clazz,name,getTypes(params));
    return executeMethod(method,target,params);
  }
}
