package de.renew.gui;

import java.util.Vector;
import de.renew.shadow.*;
import de.renew.simulator.*;
import CH.ifa.draw.framework.DrawingEditor;

public class CPNSimulation {
  private ShadowNetSystem netSystem;
  private Vector drawings = new Vector();
  private Simulator simulator = null;
  private FigureLookup figureLookup = null;

  public CPNSimulation(ShadowCompiler compiler) {
    netSystem=new ShadowNetSystem(compiler);
  }

  // Only called from CPNDrawing!
  void addDrawing(CPNDrawing drawing) {
    synchronized(drawings) {
      drawings.addElement(drawing);
    }
  }

  // Only called from CPNDrawing!
  void releaseDrawing(CPNDrawing drawing) {
    synchronized(drawings) {
      drawings.removeElement(drawing);
    }
    drawing.discardShadow();
  }

  CPNDrawing getDrawing(String name) {
    if (name==null) {
      return null;
    }
    synchronized(drawings) {
      for (int i=0;i<drawings.size();i++) {
	CPNDrawing currentDrawing=(CPNDrawing)drawings.elementAt(i);
	String currentName=currentDrawing.getName();
	if (name.equals(currentName)) {
	  return currentDrawing;
	}
      }
    }
    return null;
  }


  public void buildAllShadows() {
    //misc.Debug.key();
    synchronized(drawings) {
      for (int i=0;i<drawings.size();i++) {
	CPNDrawing currentDrawing=(CPNDrawing)drawings.elementAt(i);
        currentDrawing.buildShadow(netSystem);
      }
    }
  }

  public ShadowNetSystem getNetSystem() {
    // Make sure that all shadows are correctly set up.
    buildAllShadows();
    // Return the shadow net system.
    return netSystem;
  }

  public void syntaxCheckOnly() throws SyntaxException {
    buildAllShadows();
    ShadowLookup shadowLookup = netSystem.compile();
    figureLookup = new FigureLookup(shadowLookup);
  }

  public String[] getLintNames() {
    return netSystem.getCompiler().getLintNames();
  }

  public void performLint(int lintNr) throws SyntaxException {
    syntaxCheckOnly();
    // Now make the check.
    netSystem.lint(lintNr); 
  }

  /**
   * Start a new simulation.
   * Return true, if the creation of the initial instance
   * was successful.
   */
  public NetInstance initSimulation(String mainNet, DrawingEditor editor)
    throws SyntaxException
  {
    simulationTerminate();
    buildAllShadows();

    ShadowLookup shadowLookup = netSystem.compile();
    figureLookup = new FigureLookup(shadowLookup);
    Net net=shadowLookup.getNet(mainNet);

    simulator=null;
    if (net != null) {
      // Create an initial net instance.
      NetInstance primaryInstance=
        de.renew.simulator.SimulatorHelper.fireInitially(net,
	  new ConsoleTracer("Initially","(0) "),true);
      if (primaryInstance!=null) {
	// Create a new simulator.
	// The actual object that is created depends
	// on the desired multiplicity. A zero multiplicity
	// denotes a sequential simulator, which is minimally
	// concurrent.
	if (simulatorMultiplicity==0) {
	  simulator=new SequentialSimulator();
	} else if (simulatorMultiplicity>=2) {
	  simulator=new ParallelSimulator(simulatorMultiplicity);
	} else {
	  simulator=new ConcurrentSimulator();
	}
      }
      return primaryInstance;
    }

    return null;
  }

  public FigureLookup getFigureLookup() { return figureLookup; }

  public boolean isSimulationActive() {
    return simulator != null && simulator.isActive();
  }

  public void simulationTerminate() {
    if (simulator!=null) {
      Simulator oldSimulator=simulator;
      simulator = null;
      oldSimulator.terminateRun();
    }
    de.renew.simulator.SearchQueue.reset();

    // The next command allows figures and thus drawings to be
    // garbage collected. Maybe this is too little too late?
    // After all, a drawing might be close before the simulation
    // is terminated. But then again, that is a dangerous
    // enterprise even without the figure lookup hanging around.
    figureLookup = null;
  }

  public void simulationStop() {
    if (simulator!=null) {
      simulator.stopRun();
    }
  }

  public void simulationRun() {
    if (simulator != null) {
      simulator.startRun();
    }
  }

  public int simulationStep() {
    if (simulator != null) {
      return simulator.step();
    } else {
      return Simulator.statusDisabled;
    }
  }

  private static int simulatorMultiplicity;

  static {
    simulatorMultiplicity =
      Integer.getInteger("de.renew.simulatorMode",1).intValue();

    if (simulatorMultiplicity<=0) {
      simulatorMultiplicity=0;
      System.err.println("Using sequential simulator ...");
    } else if (simulatorMultiplicity>=2) {
      System.err.println("Using "+simulatorMultiplicity+
        " parallel simulators ...");
      System.err.println("Caution! This is an experimental feature!");
    }
  }
}
