package de.renew.simulator;

import collections.*;
import de.renew.unify.*;
import de.renew.util.*;

class ArcOccurrence implements ExecutableOccurrence {
  PlaceInstance placeInstance;
  VariableMapper mapper;
  Variable variable;
  Arc arc;

  ArcOccurrence(Arc arc, VariableMapper mapper, NetInstance netInstance) {
    this.arc=arc;
    this.mapper=mapper;
    placeInstance=netInstance.getInstance(arc.place);
  }

  public CollectionEnumeration makeBinders(Searcher searcher)
    throws Impossible 
  {
    // For the moment, arc expressions are always early.
    // We might make them late for output arcs, but this
    // might cause significant confusion. By introducing a
    // new variable and an action statement a late evaluation
    // can always be enforced within the model.
    variable=new Variable(
      arc.expression.startEvaluation(mapper,searcher.recorder,
        searcher.calcChecker),
      searcher.recorder);

    // Prepare a sequence of binders for this arc.
    UpdatableSeq binders=new LinkedList();

    if (arc.arcType==Arc.out) {
      // We must make sure that the variable associated with
      // an output arc is going to be computed before we declare
      // a transition enabled.
      searcher.calcChecker.addLateVariable(variable,
        searcher.recorder);
    } else {
      // Input arcs and double arcs simply check the
      // existance of tokens during binding, so we can use
      // identical binders for them. Test arcs
      // require a different treatment, though.
      TokenLister lister;
      if (arc.expression.isInvertible()) {
	if (arc.isTestArc()) {
	  lister=new TestTokenLister(placeInstance);
	} else {
	  lister=new InputTokenLister(placeInstance);
	}
      } else {
	lister=null;
      }

      TokenCollector collector;
      if (arc.isTestArc()) {
	collector=new TestTokenCollector(searcher.reserver,placeInstance);
      } else {
	collector=new InputTokenCollector(searcher.reserver,placeInstance);
      }
      
      binders.insertFirst(new InputPlaceBinder(variable,lister,collector));
    }

    return binders.elements();
  }

  public CollectionEnumeration makeExecutables(Copier copier,
    VariableMapper copiedMapper)
  {
    // There will be no undo past this point, so we need
    // not supply a state recorder. However, we must always
    // make a copy, because the current bindings will be
    // completely rolled back. Even when we make a getValue()
    // call to the old variable, that variable might contain
    // a tuple that is taken apart later on.
    Variable copiedVariable=(Variable)copier.copy(variable);

    // Depending on the type of the arc we must create different
    // executables.
    switch (arc.arcType) {
    case Arc.in:
      return new SingletonEnumeration(
        new InputArcExecutable(placeInstance,
          copiedVariable.getValue(),false,arc.getTrace()));
    case Arc.test:
      return new PairEnumeration(
        new TestArcExecutable(placeInstance,
          copiedVariable.getValue(),false,arc.getTrace()),
        new UntestArcExecutable(placeInstance,
          copiedVariable,arc.getTrace()));
    case Arc.out:
      return new SingletonEnumeration(
        new OutputArcExecutable(placeInstance,
          copiedVariable,arc.getTrace()));
    case Arc.both:
      return new PairEnumeration(
	new InputArcExecutable(placeInstance,
          copiedVariable.getValue(),false,arc.getTrace()),
        new OutputArcExecutable(placeInstance,
          copiedVariable,arc.getTrace()));
    case Arc.fastBoth:
      return new SingletonEnumeration(
        new InputArcExecutable(placeInstance,
          copiedVariable.getValue(),true,arc.getTrace()));
    case Arc.fastTest:
      return new SingletonEnumeration(
        new TestArcExecutable(placeInstance,
          copiedVariable.getValue(),true,arc.getTrace()));
    default:
      throw new RuntimeException("Bad arc type.");
    }
  }
}
