package de.renew.simulator;

import collections.*;
import de.renew.unify.*;
import de.renew.util.Lock;
import de.renew.util.Null;
import java.util.Enumeration;
import java.util.Vector;

// I will collect tokens for a single place instance. I also
// keep count of the number of tokens that are currently being tested.
//
// Whenever my marking is changed I will notify my searchable collection.
//
// Anybody who needs to make sure that my marking does not change should
// lock on me. Synchronisation does not help!

public class PlaceInstance {
  private NetInstance netInstance;
  Place place;

  private TokenBag freeTokens;
  private TokenBag testedTokens;
  private UpdatableSeq initTokens;

  private TupleIndex freeIndex;
  private TupleIndex testIndex;

  private SearchableCollection searchables;

  public final long lockOrder;
  public final Lock lock;

  private Vector observers;

  PlaceInstance(NetInstance netInstance, Place place)
    throws Impossible
  {
    this.netInstance=netInstance;
    this.place=place;

    lockOrder=de.renew.util.Orderer.getTicket();
    lock=new Lock();

    freeTokens=new TokenBag();
    testedTokens=new TokenBag();

    freeIndex=new TupleIndex();
    testIndex=new TupleIndex();

    searchables=new SearchableCollection();

    observers=new Vector();

    // Let's calculate the initial bag of tokens.
    initTokens=new LinkedList();
    CollectionEnumeration inscriptions=place.inscriptions.elements();
    while (inscriptions.hasMoreElements()) {
      InitialMarking initialMarking=
	(InitialMarking)inscriptions.nextElement();
      Object value=
	initialMarking.getInitialMarking();
      initTokens.insertFirst(Null.convert(value));
    }
  }

  public String toString() {
    return netInstance.toString()+"."+place.toString();
  }

  // Add a searchable. We do not lock the place instance during
  // accesses to the searchables contrary to earlier implementations.
  // The searchables have to take care of locking themselves.
  public SearchableCollection searchables() {
    return searchables;
  }

  // I should rather return a copy of
  // my token bag. This is dangerous !!! 
  public TokenBag getTokens() {
    return freeTokens;
  }

  public CollectionEnumeration getDistinctTokens() {
    return freeIndex.getAllElements();
  }

  public CollectionEnumeration getDistinctTokens(Object pattern) {
    return freeIndex.getPossibleMatches(pattern);
  }

  public CollectionEnumeration getDistinctTestableTokens() {
    return testIndex.getAllElements();
  }

  public CollectionEnumeration getDistinctTestableTokens(Object pattern) {
    return testIndex.getPossibleMatches(pattern);
  }

  public int getNumberOfTestedTokens() {
    // We do not count multiple tests on a single token as
    // multiple tokens.
    return testedTokens.getUniqueSize();
  }

  boolean containsToken(Object token) {
    return freeTokens.includes(token);
  }
  // Here consider only already tested tokens.
  public boolean containsTestedToken(Object token) {
    return testedTokens.includes(token);
  }
  // Here we consider both kinds of tokens.
  boolean containsTestableToken(Object token) {
    // nulls will be converted by the called methods.
    return containsToken(token) || containsTestedToken(token);
  }

  void removeToken(Object token) {
    freeTokens.removeOneOf(token);

    if (!freeTokens.includes(token)) {
      freeIndex.remove(token);
      if (!testedTokens.includes(token)) {
	testIndex.remove(token);
      }
    }

    searchables.proposeSearch(false,true);
    markingChanged();
  }

  // Multiple tests on the same token do not remove
  // the token multiple times. Instead, the token is
  // removed once, and the number of removals is recorded
  // by a multiple insertion into the bag of tested tokens.
  void testToken(Object token) {
    if (!testedTokens.includes(token)) {
      freeTokens.removeOneOf(token);
      if (!freeTokens.includes(token)) {
	freeIndex.remove(token);
      }

      // We must notify the listeners only if a free token has
      // become tested. If a token was already tested, the new
      // test simply delays the time span until a possible
      // activation.
      searchables.proposeSearch(false,true);
      markingChanged();
    }

    testedTokens.add(token);
    // No need to change the test index, because a free token
    // has become testable or there was a testable token already.
  }

  private void silentlyInsertToken(Object token) {
    if (!freeTokens.includes(token)) {
      freeIndex.insert(token);
      // Maybe there is an appropriate token being tested?
      if (!testedTokens.includes(token)) {
	// No, remember the new token.
	testIndex.insert(token);
      }
    }

    freeTokens.add(token);
  }

  void insertToken(Object token) {
    silentlyInsertToken(token);
    searchables.proposeSearch(true,false);
    markingChanged();
  }

  // Undo a testToken call. Once the last token is returned,
  // a token is reinserted into the bag of free tokens.
  void untestToken(Object token) {
    testedTokens.removeOneOf(token);
    // No need to change the test index, because a tested token
    // will become free or an equivalent token will remain tested.

    if (!testedTokens.includes(token)) {
      if (!freeTokens.includes(token)) {
	freeIndex.insert(token);
      }
      freeTokens.add(token);

      // We must notify the listeners only if a tested token has
      // become free. If a token is still tested, the 
      // untest simply shortens the time span until a possible
      // activation.
      searchables.proposeSearch(true,false);
      markingChanged();
    }
  }

  /**
   * My net will call this method if its creation is confirmed.
   * I will free all tokens that I kept during my initialisation.
   * You must call this method at most once.
   **/
  void createConfirmation(Tracer tracer) {
    // The following code is synchronized to avoid concurrent
    // update requests. This does not preclude all conflicts
    // that such a notification might cause. The observers must
    // still take some care although it might work now.
    // NOTE: CAREFUL EXAMINATION REQUIRED!!!
    synchronized (observers) {
      CollectionEnumeration enum=initTokens.elements();
      // Anything to do?
      if (enum.hasMoreElements()) {
	// Yes, move all tokens to their destination.
	do {
	  Object token=Null.convert(enum.nextElement());
	  if (place.trace) {
            tracer.trace("    Putting "+token+
              " into "+this);
	  }
	  silentlyInsertToken(token);
	} while (enum.hasMoreElements());
	
	// Just notify the observers once.
	searchables.proposeSearch(true,false);
	markingChanged();
      }
      
      // Null the field to allow garbage collection.
      initTokens=null;
    }
  }

  // I used to extend Obserable, but unfortunately,
  // this class is a memory sink until Java 1.1 (internal
  // array arr is not nulled). I reimplement the essential
  // functionality.
  public void addObserver(NetObserver o) {
    synchronized (observers) {
      if (!observers.contains(o)) {
	observers.addElement(o);
      }
    }
  }

  public void deleteObserver(NetObserver o) {
    synchronized (observers) {
      observers.removeElement(o);
    }
  }

  // Notify the observers that a token move was performed.
  private void markingChanged() {
    // The following code is synchronized to avoid concurrent
    // update requests. This does not preclude all conflicts
    // that such a notification might cause. The observers must
    // still take some care although it might work now.
    // NOTE: CAREFUL EXAMINATION REQUIRED!!!
    //
    // This works quite contrary to the observable class, where
    // the observable is explicitly not synchronized
    // during notification!
    synchronized (observers) {
      int size=observers.size();
      for (int i=0;i<size;i++) {
	NetObserver observer=(NetObserver)observers.elementAt(i);
	observer.update(this,freeTokens);
      }
    }
  }
}

