package de.renew.simulator;

import collections.*;

/**
 * The class search queue keeps track of the transitions
 * instances that must be searched for activated bindings.
 * This is done statically, so that a net instance
 * can register its transitions as possibly activated even if it
 * does not know of any specific searcher.
 * A searcher can ask the search queue for
 * possibly activated transitions.
 *
 * In theory, this might also be incorporated in the Searcher
 * class, but someday there might be multiple searchers
 * acting concurrently and it would be quite confusing if the
 * different searchers were synchronized via a static search queue
 * inside the searcher class. By separating this class,
 * it gets clear that a single common search queue has been implemented.
 * In the sense of load balancing it is also sensible to have a central
 * search queue.
 *
 * @author Olaf Kummer
 **/

public abstract class SearchQueue {
  private static SearchQueueData instance=new RandomSearchQueue();
  private static UpdatableSeq listeners=new LinkedList();

  // Set the instance of SearchQueueData that manages the
  // queue policy. This eanbles us to change search queue
  // policies at runtime under the control of the simulator.
  public synchronized static void setQueue(SearchQueueData queue) {
    queue.init(instance.elements());
    instance=queue;
  }

  // Clear all pending enabled transition instances in the queue.
  // This is required when a new simulation is started. Note that
  // there might be threads running in the background that
  // might reintroduce transition instances into the queue.
  // So be sure to kill everything in sight before calling
  // this method.
  public synchronized static void reset() {
    try {
      // Try to create search queue of the same type first.
      instance=(SearchQueueData)instance.getClass().newInstance();
    }
    catch (Exception e) {
      // Create a default search queue.
      instance=new RandomSearchQueue();
    }
  }

  public synchronized static void insertListener(
    SearchQueueListener listener)
  {
    listeners.insertFirst(listener);
  }
  
  protected synchronized static void notifyListeners() {
    CollectionEnumeration enum=listeners.elements();
    while (enum.hasMoreElements()) {
      ((SearchQueueListener)enum.nextElement()).searchQueueNonempty();
    }
    listeners.clear();
  }

  // Forward calls to the instance. This is required
  // so that we can already obtain a lock that will allow
  // a clean switching of instances.
  public synchronized static void include(
    Searchable searchable)
  {
    // Clear all triggers. The searchable will be inserted
    // into the search queue, so there is no danger that
    // its enabledness will not be detected. This has to
    // be done regardless of the policy of the search queue.
    searchable.triggers().clear();

    // Do it.
    instance.include(searchable);

    // Does anybody want a searchable?
    notifyListeners();
  }

  public synchronized static boolean isEmpty() {
    return instance.isEmpty();
  }

  public synchronized static Searchable extract() {
    return instance.extract();
  }
}
