package de.renew.simulator;

import collections.*;

// Recently added synchronisation. Will be necessary
// in concurrent environments. Should be covered by
// the place locking scheme, but let's not take any risks,
// because this might change.

/*

The general picture: *Before* relying on a marking for the analysis
of a searchable's enabledness, the searchable is registered with
the place's searchable collection.

*After* a change is done to a place, the place's searchable collection
is used to inform all transitions of their possible enabledness.

Each transition removes itself from *all* its search triggers.
*Afterwards* it inserts itself into the search queue.

This ensures that each marking change is guaranteed to force the
reexamination of the affected transitions. Transitions might
be examined too often, but only when a change was made
during their examination.

Lock order for deadlock exclusion:
  1. searchable collections
  2. trigger collections
The locking is done by synchronisation.

The global data structure:
Each searchable has an associated trigger collection.
Searchable collections reference searchables and make
sure that they are referenced by the associated trigger collections.

The associations between trigger collections and searchable collections
cannot be stored in a central data structure, because this would
eliminate garbage collection. 

Because there are now links in both directions, a searchable
references its triggers even after the tokens that gave rise to the
possible connection have disappeared. Even if a black token
is removed from the input of a searchable, that searchable needs 
to be rechecked, because it might be determined that a
triggering is no longer possible..

 */

public class SearchableCollection {
  // Keep track of notified objects.
  private UpdatableSet searchables;

  public SearchableCollection() {
    searchables=new HashedSet();
  }

  public synchronized void include(Searchable searchable) {
    TriggerCollection triggers=searchable.triggers();
    synchronized(triggers) {
      searchables.include(searchable);
      triggers.include(this);
    }
  }

  public synchronized void exclude(Searchable searchable) {
    TriggerCollection triggers=searchable.triggers();
    synchronized(triggers) {
      searchables.exclude(searchable);
      triggers.exclude(this);
    }
  }

  public void proposeSearch(boolean mightActivate, boolean mightDeactivate) {
    LinkedList list=new LinkedList();
    synchronized(this) {
      // Get the searchables.
      list.prependElements(searchables.elements());
    }

    // Send the notifications. No need to synchronize here.
    // Better allow concurrent accesses to reduce deadlock potential.
    // This way the searchables can exclude themselves during the
    // notification.
    CollectionEnumeration enum=list.elements();
    while (enum.hasMoreElements()) {
      Searchable searchable=(Searchable)enum.nextElement();
      searchable.proposeSearch(mightActivate,mightDeactivate);
    }
  }
}
