package de.renew.simulator;

import java.lang.reflect.InvocationTargetException;
import de.renew.unify.*;

public class SequentialSimulator
  implements Simulator, Runnable
{
  public static void main(String[] args) {
    if (args.length==0) {
      System.out.println(
        "Usage: java de.renew.simulator.SequentialSimulator [CLASS]");
      System.exit(0);
    }

    // Create a net.
    Net net=null;
    try {
      net=(Net)Class.forName(args[0]).
        getMethod("instance",new Class[0]).invoke(null,new Object[0]);
    }
    catch (Exception e) {
      System.out.println(args[0]+
        " isn't a valid net class as it seems.");
      if (e instanceof InvocationTargetException) {
	((InvocationTargetException)e).getTargetException().printStackTrace();
      }
      e.printStackTrace();
      System.exit(1);
    }
    
    
    NetInstance primaryInstance=SimulatorHelper.fireInitially(net,
      new ConsoleTracer("Initially","(0) "),true);

    if (primaryInstance!=null) {
      Simulator simulator=new SequentialSimulator();
      simulator.startRun();
    } else {
      System.out.println("The net did not accept the init request.");
    }
  }

  private ExecuteFinder finder;
  private Searcher searcher=new Searcher();
  private boolean stepFired=false;

  private boolean wantBreak=false;
  private Thread runThread=null;

  public SequentialSimulator() {
    // If possible, prepare a new binding.
    findNewBinding();
  }

  public boolean isActive() {
    return finder.isExecutable();
  }

  public synchronized void startRun() {
    // Only create a new thread if none is currently running.
    if (runThread==null) {
      // Create a new thread that will run the simulation.
      runThread=new Thread(this);
      // Run the thread in the background.
      runThread.setPriority(Thread.MIN_PRIORITY);
      // Start the thread.
      runThread.start();
    }
  }

  // Gently stop the simulation.
  // This may take a while, because the current search for
  // a activated binding must be completed first.
  // After the completion of this method, no thread is running
  // and runThread==null.
  public synchronized void stopRun() {
    while (runThread!=null) {
      wantBreak=true;
      try {
	wait();
      }
      catch (InterruptedException e) {
	// This is expected. But I cannot be sure that
	// I woke up from a notification.
      }
    }
  }


  // Terminate the simulation once and for all.
  // Do some final clean-up and exit the thread.
  public synchronized void terminateRun() {
    // In this case we can simply call stop run, because that
    // is sufficient to terminate all running threads.
    stopRun();
  }

  private void fire(Tracer tracer) {
    stepFired=finder.isExecutable();
    if (stepFired) {
      finder.execute(tracer,false);
      findNewBinding();
    }
  }

  // To exclude concurrent access to the searcher, this method may only
  // be called from the running thread or from a
  // synchronized method that has ensured that no thread is running.
  // This method is not synchronized, because it could
  // lock up termination requests.
  // If there are no more enabled transitions, this method simply
  // returns without effect.
  private void findNewBinding() {
    // An ExecuteFinder should not be recycled.
    finder=new ExecuteFinder();
    // Let's look for another binding.
    searcher.searchLoop(finder);
  }

  // This method must be synchronized to avoid concurrent
  // accesses to restart another thread after the stop.
  public synchronized int step() {
    if (runThread!=null) {
      // Just stop the current thread. It will make another step
      // anyway.
      stopRun();
    } else {
      // Try to make another step.
      fire(new ConsoleTracer("Synchronously"));
    }
    if (stepFired) {
      if (finder.isExecutable()) {
	return statusStepComplete;
      } else {
	return statusLastComplete;
      }
    } else {
      return statusDisabled;
    }
  }

  // This method is called in threads that are generated on each
  // run request. The simulator class cannot inherit from thread,
  // because this method might be run multiple times.
  public void run() {
    // Do have a prepared binding?
    if (!finder.isExecutable()) {
      // No, let's try to find one.
      findNewBinding();
    }
    // Make steps until deadlock or until a break is requested.
    while (finder.isExecutable() && !wantBreak) {
      // Grant other threads a chance to run.
      // Only needed on systems without preemptive multitasking.
      Thread.yield();
      // Fire and search for a new binding.
      fire(new ConsoleTracer("Synchronously"));
    }
    // Make sure that runThread and wantBreak are updated
    // atomically.
    synchronized(this) {
      // Signal the end of the run.
      runThread=null;
      // Clear the break request.
      wantBreak=false;
      // Notify waiting threads of break.
      notifyAll();
    }
  }
}
