package de.renew.simulator;

import collections.*;
import de.renew.unify.*;

public class SimulatorHelper {
  public static NetInstance fireInitially(Net net,
    Tracer tracer, boolean asynchronous)
  {
    // Search for the initial transition and fire it.
    // We make sure that the primary instance exists
    // only after a successful initial synchronisation.
    // We create a searcher especially for this purpose.
    // This is not overly efficient, but we do not expect
    // to fire a primary instance often.
    Searcher searcher=new Searcher();

    // Create the primary instance.
    NetInstance primaryInstance;
    try {
      primaryInstance=net.makeInstance();
    }
    catch (Impossible e) {
      primaryInstance=null;
    }

    // If the creation of the primary instance succeeded,
    // Try to perform the initial synchronisation.
    ExecuteFinder finder=new ExecuteFinder();
    if (primaryInstance!=null) {
      // Fire the primary transition to get an
      // initial marking. The parameters could be
      // dependent on the parameters of this method.
      // No need to undo this action.
      Variable variable=new Variable(new Tuple(0),null);
      
      // Prepare the execution of the net. If no channel with
      // name "new" is present, simply ignore the synchronisation.
      searcher.initiatedSearch(primaryInstance,
	"new",variable,true,finder);

      // Did we find a binding?
      if (!finder.isExecutable()) {
	primaryInstance=null;
      }
    }

    // Execute the binding.
    if (primaryInstance!=null) {
      // Confirm the permanent activation for the
      // fully spontaneous transitions of this net. The
      // confirmation is only done if the synchronisation succeeds,
      // because conceptionally, the net instance is not created
      // otherwise. But the confirmation must be done before the
      // initial transition is executed, because another binding
      // is sought immediately.
      primaryInstance.createConfirmation(tracer);
      
      // Execute the initial transition. Its binding
      // is recorded in the finder.
      finder.execute(tracer,asynchronous);
    }

    return primaryInstance;
  }

  public static boolean isFirable(TransitionInstance instance, 
    Searcher searcher)
  {
    EnablednessFinder finder=new EnablednessFinder();
    searcher.searchOnce(finder,instance);
    return finder.isEnabled();
  } 

  public static boolean isFirable(TransitionInstance instance) {
    Searcher searcher=new Searcher();
    return isFirable(instance,searcher);
  } 

  public static boolean fireOneBinding(TransitionInstance instance, 
    Tracer tracer, boolean asynchronous, Searcher searcher)
  {
    ExecuteFinder finder=new ExecuteFinder();
    searcher.searchOnce(finder,instance);
    if (finder.isExecutable()) {
      finder.execute(tracer,asynchronous);
      return true;
    }
    return false;
  } 

  public static boolean fireOneBinding(TransitionInstance instance,
    Tracer tracer, boolean asynchronous)
  {
    Searcher searcher=new Searcher();
    return fireOneBinding(instance,tracer,asynchronous,searcher);
  } 

  public static CollectionEnumeration findAllBindings(
    TransitionInstance instance, Searcher searcher)
  {
    CollectingFinder finder=new CollectingFinder();
    searcher.searchOnce(finder,instance);
    return finder.bindings();
  } 

  public static CollectionEnumeration findAllBindings(
    TransitionInstance instance)
  {
    Searcher searcher=new Searcher();
    return findAllBindings(instance,searcher);
  } 

  // The lock method must be called first. If it fails with
  // an exception, the transition cannot fire. If it succeeds,
  // it must be followed by a call to executeEarly() and then
  // by a call to executeLate().
  private static void lock(final Seq executables) throws Impossible {
    UpdatableSeq unlockables=new LinkedList();

    // Lock all early executables.
    try {
      CollectionEnumeration enum=executables.elements();
      while (enum.hasMoreElements()) {
        Executable executable=(Executable)enum.nextElement();
        if (executable instanceof EarlyExecutable) {
	  ((EarlyExecutable)executable).lock();
	  // Ok, I got the lock. Now I must make sure to
	  // treat this executable correctly: unlock or execute.
	  unlockables.insertFirst(executable);
        }
      }
    }
    catch (Impossible e) {
      // Something failed. Probably a missing token.
      CollectionEnumeration enum=unlockables.elements();
      while (enum.hasMoreElements()) {
	((EarlyExecutable)enum.nextElement()).unlock();
      }
      // I unlocked everything (in reverse order, if that matters).
      // Now I propagate the orginal exception to my caller.
      throw e;
    }
  }

  static void executeEarly(final Seq executables, 
    final Tracer tracer)
  {
    // All locks were correctly granted. Now we call all early
    // executables in their appropriate order. The execute method
    // will automatically release the executable's lock on exit.

    CollectionEnumeration enum=executables.elements();
    while (enum.hasMoreElements()) {
      Executable executable=(Executable)enum.nextElement();
      if (executable instanceof EarlyExecutable) {
	((EarlyExecutable)executable).execute(tracer);
      }
    }
  }

  static void executeLate(final Seq executables, 
    final Tracer tracer)
  {
    // All early executables were correctly executed and
    // the locks were released. Now we call all late
    // executables in their appropriate order. We do this
    // in a separate thread to avoid locking up the simulation.

    CollectionEnumeration lateEnum=executables.elements();
    while (lateEnum.hasMoreElements()) {
      Executable executable=(Executable)lateEnum.nextElement();
      if (executable instanceof LateExecutable) {
	if (((LateExecutable)executable).isLong()) {
	  de.renew.util.Detacher.detach();
	}
	try {
	  executable.execute(tracer);
	}
	catch (Exception e) {
	  tracer.trace("Exception occurred during late phase of firing:");
	  tracer.trace(e.toString());
	  e.printStackTrace();
	  // Do not proceed. Some variables might have remained
	  // unbound, so it is unlikely that the following
	  // executables will succeed.
	  tracer.complete();
	  return;
	}
      }
    }
    // Everything is finished. Let me just inform the tracer.
    tracer.complete();
  }

  public static boolean execute(final Seq executables, 
    final Tracer tracer, final boolean asynchronous)
  {
    // Execute the binding that is already stored in the finder.
    try {
      lock(executables);
      // Ok, no exception was thrown.
      executeEarly(executables,tracer);
      // The remaining steps can be done in the background,
      // if we want to do so.
      if (asynchronous) {
	de.renew.util.Detacher.startAttached(new Thread(
          new LateRunnable(executables,tracer)));
      } else {
	executeLate(executables,tracer);
      }
      
      return true;
    }
    catch (Impossible e) {
      // The transition could not fire. Probably
      // a token disappeared in the mean time.
      // System.out.println("The calculated binding was disabled.");
      // System.out.println("Searching for an alternative binding.");

      return false;
    }
  }
}
