package de.renew.simulator;

import collections.*;
import de.renew.util.Null;

// A token reservation keeps track of the number of tokens
// that would have to be tested and removed from a single
// place instance.
//
// Accesses to objects of this class are unsynchronized.
// It is assumed that any access is guarded by a lock on the
// associated token reserver.

class TokenReservation {
  private PlaceInstance instance;
  private int reservations;

  private TokenBag testedTokens;
  private TokenBag removedTokens;

  TokenReservation(PlaceInstance instance) {
    this.instance=instance;
    reservations=0;
    
    testedTokens=new TokenBag();
    removedTokens=new TokenBag();
  }

  boolean isRemovable() {
    return reservations==0;
  }

  public boolean containsRemovableToken(Object token) {
    instance.lock.lock();

    int required=removedTokens.getMultiplicity(token)+1;
    boolean isTested=testedTokens.getMultiplicity(token)>0;
    if (isTested && !instance.containsTestedToken(token)) {
      required++;
    }
    boolean result=instance.getTokens().getMultiplicity(token)>=required;

    instance.lock.unlock();
    return result;
  }

  public boolean containsTestableToken(Object token) {
    instance.lock.lock();

    int required=removedTokens.getMultiplicity(token);
    if (!instance.containsTestedToken(token)) {
      required++;
    }
    boolean result=instance.getTokens().getMultiplicity(token)>=required;

    instance.lock.unlock();
    return result;
  }

  public boolean removeToken(Object token) {
    instance.lock.lock();

    boolean result=containsRemovableToken(token);
    if (result) {
      removedTokens.add(token);
      reservations++;
    }

    instance.lock.unlock();
    return result;
  }

  public boolean testToken(Object token) {
    instance.lock.lock();

    boolean result=containsTestableToken(token);
    if (result) {
      testedTokens.add(token);
      reservations++;
    }

    instance.lock.unlock();
    return result;
  }

  public void unremoveToken(Object token) {
    removedTokens.removeOneOf(token);
    reservations--;
  }

  public void untestToken(Object token) {
    testedTokens.removeOneOf(token);
    reservations--;
  }

  public CollectionEnumeration getRemovableTokens(Object pattern) {
    instance.lock.lock();

    CollectionEnumeration tokens=instance.getDistinctTokens(pattern);
    UpdatableSeq result=new LinkedList();
    while (tokens.hasMoreElements()) {
      Object token=tokens.nextElement();
      if (containsRemovableToken(token)) {
	result.insertFirst(Null.convert(token));
      }
    }

    instance.lock.unlock();
    return Null.convertEnumeration(result.elements());
  }

  public CollectionEnumeration getTestableTokens(Object pattern) {
    instance.lock.lock();

    CollectionEnumeration tokens=instance.getDistinctTestableTokens(pattern);
    UpdatableSeq result=new LinkedList();
    while (tokens.hasMoreElements()) {
      Object token=tokens.nextElement();
      if (containsTestableToken(token)) {
	result.insertFirst(Null.convert(token));
      }
    }

    instance.lock.unlock();
    return Null.convertEnumeration(result.elements());
  }
}

