package de.renew.simulator;

import collections.*;

/**
 * I represent a transition with all its semantic information, but
 * without any state. I can have a name, an associated net, an
 * uplink, and a set of inscriptions. My main job is to store
 * all this information. I do not actaully perform any
 * useful services.
 **/
public class Transition {
  /**
   * My net.
   **/
  final Net net;
  /**
   * My name.
   **/
  final String name;

  /**
   * true, if my instances should output a trace message when they
   * occur. This is the default.
   **/
  boolean trace;

  /**
   * My uplink or null, if I am spontaneous.
   **/
  UplinkInscription uplink;
  
  /**
   * A set of inscriptions. I expect all inscription objects to be
   * of the type TransitionInscription.
   *
   * @see de.renew.simulator.ActionInscription
   * @see de.renew.simulator.Arc
   * @see de.renew.simulator.CreationInscription
   * @see de.renew.simulator.DownlinkInscription
   * @see de.renew.simulator.ExpressionInscription
   **/
  UpdatableSet inscriptions;

  /**
   * I (a transition) am created as a spontaneous transition.
   * Later on, inscriptions may be added, even an uplink.
   * I am automatically registered at the net as a
   * net element.
   *
   * @param net the net to which this transition should be added.
   **/
  public Transition(Net net, String name) {
    // Remember the associated net.
    this.net=net;
    this.name=name;

    // Trace actions.
    trace=true;

    // Clear the uplink: no uplink is the default.
    uplink=null;

    // Initialize the array for the inscriptions.
    inscriptions=new HashedSet();

    // Notify the net that we are done and want to be
    // inserted into the list of transitions.
    net.add(this);
  }

  /**
   * I return my name as my string representation.
   * My net instances use my name as a prefix for their name.
   * @return my name.
   **/
  public String toString() {
    return name;
  }

  /**
   * Switch my trace flag on or off.
   *
   * @param trace true, if tracing is desired
   **/
  public void setTrace(boolean trace) {
    this.trace=trace;
  }

  /**
   * Am I being traced?
   *
   * @return true, if my trace flag is set
   **/
  public boolean getTrace() {
    return trace;
  }

  /**
   * Am I spontaneous?
   *
   * @return true, if I am spontaneous, i.e., if I do not
   *   have an uplink
   **/
  public boolean isSpontaneous() {
    return uplink==null;
  }

  /**
   * Here I extract my uplink, if any, from my inscriptions.
   * If there is more than one uplink, it is undetermined which
   * uplink I will use. This method is called after every
   * update to my set of inscriptions.
   **/
  private void checkUplink() {
    // Maintain the count of uplinks and select an uplink
    // from the set of inscriptions. By allowing multiple inscriptions
    // the embedding into a graphical editor becomes easier.
    // On the other hand, the transition might get into
    // an inconsistent state. We try to make it as simulatable as
    // possible, however.
    uplink=null;
    CollectionEnumeration enum=inscriptions.elements();
    while (enum.hasMoreElements()) {
      Object inscription=enum.nextElement();
      if (inscription instanceof UplinkInscription) {
	uplink=(UplinkInscription)inscription;
      }
    }
  }

  /**
   * Give me another inscription. After inserting
   * the inscriptions, I will recheck my uplink.
   **/
  public void add(TransitionInscription transitionInscription) {
    if (!inscriptions.includes(transitionInscription)) {
      inscriptions.include(transitionInscription);
      checkUplink();
    }
  }
  
  /**
   * Remove one my current inscriptions. Check if an
   * uplink is still provided.
   **/
  public void remove(TransitionInscription transitionInscription) {
    if (inscriptions.includes(transitionInscription)) {
      inscriptions.exclude(transitionInscription);
      checkUplink();
    }
  }
}

