package de.renew.simulator;

import collections.*;
import de.renew.unify.Impossible;

/**
 * I represent an instance of a transition. Unlike a naked transition
 * I am tied to a net instance. I provide my name to the outside world
 * and keep track of those objects who like to be notified if I become
 * enabled.
 *
 * Other objects will register me as a listener of objects that might
 * enable me. I would like to be registered as a listener with
 * all objects that might enable me, but this is beyond my control.
 * So take a little care when you use me.
 *
 * I used to send out messages in the case that I might become
 * enabled. This functionality was removed, because no one
 * really ever needed it.
 *
 * @author Olaf Kummer
 **/

public class TransitionInstance implements Searchable {
  /**
   * My net instance to which I belong.
   **/
  NetInstance netInstance;
  /**
   * The transition that serves as a pattern for me.
   **/
  Transition transition;
  /**
   * The triggers that might notify me.
   **/
  private TriggerCollection triggers=new TriggerCollection(this);

  /**
   * I (a transition instance) am created. I remember
   * the neccessary information.
   *
   * @param netInstance
   *   the net instance that needs to have an instance of some transition
   * @param transition
   *   the transition that must be instantiated
   **/
  TransitionInstance(NetInstance netInstance,
    Transition transition)
  {
    this.netInstance=netInstance;
    this.transition=transition;
  }

  /**
   * I will use the name of my net instance plus the name of
   * my transition to build my own name.
   *
   * @return my name
   **/
  public String toString() {
    return netInstance.toString()+"."+transition.toString();
  }

  public TriggerCollection triggers() {
    return triggers;
  }

  /**
   * Somebody will call this method to tell me that I might
   * be enabled. I will tell the search queue about it, if I am
   * spontaneous.
   */
  public void proposeSearch(boolean mightActivate, boolean mightDeactivate) {
    // Notify the searcher, if this transition is
    // spontaneous. Otherwise, the notification was sporadic.
    // Maybe I should report an error instead of silently
    // ignoring the notification.
    if (transition.isSpontaneous()) {
      SearchQueue.include(this);
    }
  }

  /**
   * My net will call this method if it creation is confirmed.
   * Just to be safe, I will register myself at the search queue
   * if I am spontaneous.
   **/
  void createConfirmation() {
    if (transition.isSpontaneous()) {
      SearchQueue.include(this);
    }
  }

  /**
   * This method is called if somebody wants me to start a
   * search for an activated binding. I initiate the search and
   * let the searcher do the rest.
   *
   * @param searcher
   *   the searcher that is supposed to actually perform the search.
   **/
  public void search(Searcher searcher) {
    try {
      // Prepare a new instance to check.
      TransitionOccurrence occurrence=
	new TransitionOccurrence(this,null,searcher);
      
      // No exception was thrown, so this transition
      // may be checked.
      searcher.addOccurrence(occurrence);
      searcher.addBinders(occurrence.getBinders());
      searcher.search();
      searcher.removeBinders(occurrence.getBinders());
      searcher.removeOccurrence(occurrence);
    }
    catch (Impossible e) {
      // I had to refuse the search request completely.
      // Probably there is a cyclic dependency of some
      // calculations. This should really be detected during
      // the syntax check.
    }

    searcher.recorder.restore(); // no bindings sensible
  }
}
