package de.renew.simulator;

import collections.*;
import de.renew.unify.*;

/**
 * Upon evaluation I will spawn the evaluation of several
 * supexpressions. I will then compose a tuple from the resulting values.
 *
 * @author Olaf Kummer
 **/
public class TupleExpression implements Expression {
  /**
   * A tuple with the expressions that I must evaluate.
   * Hmm, honestly, is there a reason why this must be
   * stored in a tuple? Inefficient. Probably that's just a
   * symmetry consideration.
   **/
  private Tuple expressions;

  /**
   * I am invertible if I have at least one invertible
   * subexpression. By default, I am supposed to be
   * invertible.
   **/
  private boolean invertible=true;

  /**
   * I (a tuple expression) am created.
   * 
   * @param expressions
   *   a tuple that holds the expressions that must be evaluated
   **/
  public TupleExpression(Tuple expressions) {
    this.expressions=expressions;
    invertible=false;
    for (int i=0;i<expressions.getArity();i++) {
      invertible=invertible ||
	((Expression)expressions.getComponent(i)).isInvertible();
    }
  }

  /**
   * Create a new tuple expression from exactly one
   * expression. Unary tuples are not used very often,
   * but they do occur sometimes.
   *
   * @param expr1
   *   the first expression
   **/
  public TupleExpression(Expression expr1) {
    this(new Tuple(new Object[]{expr1},null));
  }

  /**
   * Create a new tuple expression from exactly two
   * expressions. Because pairs are so common, it seems
   * desirable to have a short notation for them.
   *
   * @param expr1
   *   the first expression
   * @param expr2
   *   the second expression
   **/
  public TupleExpression(Expression expr1, Expression expr2) {
    this(new Tuple(new Object[]{expr1,expr2},null));
  }

  /**
   * Create a new tuple expression from exactly two
   * expressions. Because triples are still farily common, it seems
   * desirable to have a short notation for them.
   *
   * @param expr1
   *   the first expression
   * @param expr2
   *   the second expression
   * @param expr3
   *   the third expression
   **/
  public TupleExpression(Expression expr1, Expression expr2,
    Expression expr3)
  {
    this(new Tuple(new Object[]{expr1,expr2,expr3},null));
  }

  /**
   * Am I invertible? I look it up in the cached field.
   *
   * @return true, if at least one of my subexpressions is invertible.
   **/
  public boolean isInvertible() {
    return invertible;
  }

  /**
   * There is really not much difference between evaluation
   * and registration. Only, I have to evaluate the subexpressions
   * accordingly.
   *
   * @param mapper
   *   a variable mapper which I must pass to the subexpressions
   * @param recorder
   *   a state recorder which I must pass to the subexpressions
   * @param checker
   *   a calculation checker which I must pass to the subexpressions
   * @param forEvaluation
   *   true, if we are trying an evaluation
   * @return the constructed tuple, possibly incomplete
   **/
  private Tuple makeResultTuple(VariableMapper mapper,
    StateRecorder recorder, CalculationChecker checker,
    boolean forEvaluation) throws Impossible
  {
    int n=expressions.getArity();

    Object[] args=new Object[n];
    for (int i=0;i<n;i++) {
      Expression expr=(Expression)expressions.getComponent(i);
      if (forEvaluation) {
 	args[i]=expr.startEvaluation(mapper,recorder,checker);
      } else {
 	args[i]=expr.registerCalculation(mapper,recorder,checker);
      }
    }

    return new Tuple(args,recorder);
  }

  /**
   * I will evaluate the subexpressions and put the results
   * into a tuple.
   *
   * @param mapper
   *   a variable mapper which I must pass to the subexpressions
   * @param recorder
   *   a state recorder which I must pass to the subexpressions
   * @param checker
   *   a calculation checker which I must pass to the subexpressions
   * @return the constructed tuple, possibly incomplete
   **/
  public Object startEvaluation(VariableMapper mapper, StateRecorder recorder, 
    CalculationChecker checker) throws Impossible
  {
    return makeResultTuple(mapper,recorder,checker,true);
  }

  /**
   * I will register the subexpressions and put the results
   * into a tuple.
   *
   * @param mapper
   *   a variable mapper which I must pass to the subexpressions
   * @param recorder
   *   a state recorder which I must pass to the subexpressions
   * @param checker
   *   a calculation checker which I must pass to the subexpressions
   * @return the constructed tuple, possibly incomplete
   **/
  public Object registerCalculation(VariableMapper mapper,
    StateRecorder recorder, CalculationChecker checker) throws Impossible
  {
    return makeResultTuple(mapper,recorder,checker,false);
  }
}

