package de.renew.unify;

import collections.*;

public final class Tuple implements Referable, Referer {
  final private BacklinkSet backlinkSet=new BacklinkSet();
  final private Reference[] references;
  final private RecorderChecker recorderChecker;

  private boolean complete; // Are no unknowns part of this object?
  private boolean bound; // Is this object bound to a real value?

  public Tuple(int arity) {
    recorderChecker=new RecorderChecker(null);
    references=Unify.makeUnknownReferenceArray(arity,this);

    complete=(arity==0);
    bound=(arity==0);
  }

  public Tuple(Object[] initValues, StateRecorder recorder) {
    recorderChecker=new RecorderChecker(recorder);
    references=Unify.cleanupReferenceArray(initValues,this,recorder);

    complete=calculateComplete();
    bound=calculateBound();
  }

  public int getArity() {
    return references.length;
  }

  public Object getComponent(int i) {
    return references[i].value;
  }

  private boolean calculateComplete() {
    for (int i=0;i<references.length;i++) {
      if (!references[i].complete) {
	return false;
      }
    }
    return true;
  }

  private boolean calculateBound() {
    for (int i=0;i<references.length;i++) {
      if (!references[i].bound) {
	return false;
      }
    }
    return true;
  }

  public boolean isComplete() {
    return complete;
  }

  public boolean isBound() {
    return bound;
  }

  void unifySilently(Tuple that, StateRecorder recorder,
    UpdatableSet listeners) throws Impossible 
  {
    if (that.references.length!=references.length) {
      throw new Impossible();
    }
    for (int i=0;i<references.length;i++) {
      Unify.unifySilently(getComponent(i),that.getComponent(i),
	recorder,listeners);
    }
  }

  public void addBacklink(Reference reference, StateRecorder recorder) {
    recorderChecker.checkRecorder(recorder);
    backlinkSet.addBacklink(reference,recorder);
  }

  public void possiblyCompleted(UpdatableSet listeners, StateRecorder recorder)
    throws Impossible
  {
    if (!complete && calculateComplete()) {
      if (recorder!=null) {
	recorder.record(new StateRestorer() {
	  public void restore() {
	    complete=false;
	  }
	});
      }
      complete=true;

      // Maybe I am even bound?
      if (calculateBound()) {
	if (recorder!=null) {
	  recorder.record(new StateRestorer() {
	    public void restore() {
	      bound=false;
	    }
	  });
	}
	bound=true;
      }
      
      // I any case, I have a notification to send to my backlinked objects.
      // I am notifying with myself as the new value.
      backlinkSet.updateBacklinked((Referable)this,this,listeners,recorder);
    }
  }

  public void occursCheck(Unknown that, UpdatableSet visited)
    throws Impossible
  {
    // Am I complete? If yes, no unknown can possibly
    // be contained within me.
    if (complete) {
      return;
    }

    // Did I check myself earlier?
    IdentityWrapper thisWrapper=new IdentityWrapper(this);
    if (visited.includes(thisWrapper)) {
      return;
    }
    // I do not want to check me again.
    visited.include(thisWrapper);

    // Now let's test all the referenced objects.
    for (int i=0;i<references.length;i++) {
      references[i].occursCheck(that,visited);
    }
  }

  public int hashCode() {
    int result=references.length+71;

    for (int i=0;i<references.length;i++) {
      if (references[i].value==null) {
	result=result*41+1;
      } else {
	result=result*41+references[i].value.hashCode();
      }
    }

    return result;
  }

  public boolean equals(Object obj) {
    if (!(obj instanceof Tuple)) {
      return false;
    }
    
    Tuple that=(Tuple)obj;
    if (references.length!=that.references.length) {
      return false;
    }

    for (int i=0;i<references.length;i++) {
      if (references[i].value==null) {
        if (that.references[i].value!=null) {
          return false;
        }
      } else {
        if (!references[i].value.equals(that.references[i].value)) {
          return false;
        }
      }
    }

    return true;
  }

  public String toString() {
    StringBuffer result=new StringBuffer();
    result.append('[');
    for (int i=0;i<references.length;i++) {
      if (i>0) {
        result.append(',');
      }
      if (references[i].value==null) {
	result.append("null");
      } else {
	result.append(references[i].value.toString());
      }
    }
    result.append(']');
    return result.toString();
  }
}
