package fs;

import collections.*;
import java.util.*;

public class ConjunctiveType implements Type {

  private UpdatableSet concepts = new HashedSet();

  protected int hashCode; // as the hashCode is a finicky computation, it is being cached

  ConjunctiveType () {
     hashCode = 0;
  }

  public ConjunctiveType (Concept concept) {
     concepts.include(concept);
     hashCode = concept.hashCode();
  }

  public int hashCode() { return hashCode; }

  /** Return an Enumeration of all Concepts this ConjunctiveType is composed 
     *  of.
     */
  public CollectionEnumeration concepts() {
    return concepts.elements();
  }

  /** Return the name of this Type. */
  public String getName() {
    StringBuffer output = new StringBuffer();
    Enumeration conceptEnum = concepts.elements();
    for (int i = 1; conceptEnum.hasMoreElements(); ++i) {
      if (i > 1)
	output.append(',');
      Concept concept = (Concept)conceptEnum.nextElement();
      output.append(concept.getName());
    }
    return output.toString();
  }

  /** Finds or constructs a ConjunctiveType consisting of the given
    * Concepts. This may fail due to incompatibilities.
    */
  public static Type getType(Set concepts) throws UnificationFailure{
    if (concepts.isEmpty())
       return Type.TOP;
    Enumeration conceptEnum = concepts.elements();
    ConjunctiveType  newtype = new ConjunctiveType();
    while (conceptEnum.hasMoreElements()) {
      newtype.addConcept((Concept)conceptEnum.nextElement());
    }
    return newtype;
   }

  public static Concept getConcept(String conceptStr) throws TypeException {
    try {
      return JavaConcept.getJavaConcept(conceptStr);
    } catch (TypeException tex) {
      return ConceptImpl.getConcept(conceptStr);
    }
  }

  public static Type getType(String conceptsStr) 
    throws UnificationFailure, TypeException {
    StringTokenizer conceptNames = new StringTokenizer(conceptsStr,",");
    HashedSet concepts = new HashedSet();
    while (conceptNames.hasMoreElements()) {
        concepts.include(getConcept(conceptNames.nextToken()));
    }
    return getType(concepts);
  }

  public boolean isExtensional() {
    Enumeration conceptEnum = concepts.elements();
    while (conceptEnum.hasMoreElements()) {
      if (((Concept)conceptEnum.nextElement()).isExtensional())
	return true;
    }
    return false;
  }

  public boolean isApprop(Integer featureCode) {
    Enumeration conceptEnum = concepts.elements();
    while (conceptEnum.hasMoreElements()) {
      if (((Concept)conceptEnum.nextElement()).isApprop(featureCode))
	return true;
    }
    return false;
  }

  public Type appropType(Integer featureCode) {
    Type typ = null;
    Enumeration conceptEnum = concepts.elements();
    while (conceptEnum.hasMoreElements()) {
      Concept concept = (Concept)conceptEnum.nextElement();
      if (concept.isApprop(featureCode)) {
	if (typ == null)
	  typ = concept.appropType(featureCode);
	else {
	  try {
	    typ = typ.unify(concept.appropType(featureCode));
	  } catch (UnificationFailure uff) {
	    return null;
	  }
	}
      }
    }
    return typ;
  }

  public CollectionEnumeration appropFeatureCodes() {
    UpdatableSet feats = new HashedSet();
    Enumeration conceptEnum = concepts.elements();
    while (conceptEnum.hasMoreElements()) {
      Enumeration featenum = ((Concept)conceptEnum.nextElement()).appropFeatureCodes();
      feats.includeElements(featenum);
    }
    return feats.elements();
  }

  public boolean subsumes(Type that) {
    Enumeration conceptEnum = concepts.elements();
  forAllConcepts:
    while (conceptEnum.hasMoreElements()) {
       Concept concept = (Concept)conceptEnum.nextElement();
       Enumeration thatConceptEnum = that.concepts();
       while (thatConceptEnum.hasMoreElements()) {
           Concept thatConcept = (Concept)thatConceptEnum.nextElement();
	   if (thatConcept.isa(concept))
                continue forAllConcepts;
        }
        return false;        // No that-concept found!
    }    
    return true;
  }


  /** Return whether this Type is subsumed by <that> Type.
    * In other words, return wheter this Type is more special than <that>
    * Type.
    */
  public boolean subsumedBy(Type that) {
     return that.subsumes(this);
  }

  public boolean canUnify(Type that) {
   Enumeration conceptEnum = concepts.elements();
    while (conceptEnum.hasMoreElements()) {
       Concept concept = (Concept)conceptEnum.nextElement();
       Enumeration thatConceptEnum = that.concepts();
       while (thatConceptEnum.hasMoreElements()) {
           Concept thatConcept = (Concept)thatConceptEnum.nextElement();
	   if (thatConcept.isNotA(concept))
                return false;
        }
    }    
    return true;
   }

  /** Adds a part Type to this ConjunctiveType. This method should
    * only be used during the construction of a new Type.
    */
  public void addConcept(Concept newConcept) throws UnificationFailure {
      hashCode = 0;
      UpdatableSet newConcepts = new HashedSet();
      Enumeration conceptEnum = concepts.elements();
      while (conceptEnum.hasMoreElements()) {
        Concept concept = (Concept)conceptEnum.nextElement();
        if (concept.isa(newConcept))
           return; // already something more special there
        if (concept.isNotA(newConcept))
           throw new UnificationFailure(); // incompatible!
        // only include concepts which are not more general than the new:
        if (!newConcept.isa(concept)) {
           newConcepts.include(concept); 
           hashCode += concept.hashCode();
        }
      }
      newConcepts.include(newConcept);
      hashCode += newConcept.hashCode();
      concepts = newConcepts;
  }

  public Type unify(Type that) throws UnificationFailure {
      ConjunctiveType  typ = new ConjunctiveType ();
      typ.concepts = (UpdatableSet)concepts.duplicate();
      Enumeration thatConceptEnum = that.concepts();
      while (thatConceptEnum.hasMoreElements()) {
          Concept thatConcept = (Concept)thatConceptEnum.nextElement();
          typ.addConcept(thatConcept);
      }
      return typ;
  }

  public Type mostGeneralExtensionalSupertype(Type that) {
     return null;
/*
//     System.out.println("Finding most general extensional supertype for "+this+" and "+that+".");
    UpdatableSet thisSupers = new HashedSet();
    Enumeration conceptEnum = concepts.elements();
    while (conceptEnum.hasMoreElements()) {
      Type type = (Concept)conceptEnum.nextElement();
      thisSupers.includeElements(type.extensionalSuperconcepts());
    }
//    System.out.println("Supertype of "+this+": "+new ConjunctiveType(thisSupers));
    UpdatableSet commonSupers = new HashedSet();
    Enumeration thatConcs = that.concepts();
    while (thatConcs.hasMoreElements()) {
      Type thattype = (Concept)thatConcs.nextElement();
      Enumeration thatSupers = thattype.extensionalSuperconcepts();
      while (thatSupers.hasMoreElements()) {
         Type thatSuper = (Type)thatSupers.nextElement();
//    System.out.print("Supertype of "+that+" "+thatSuper.getName());
         if (thisSupers.includes(thatSuper)) {
//            System.out.println(" is added!");
            if (!(new ConjunctiveType(commonSupers).subsumes(new ConjunctiveType(thatSuper)))) {
               UpdatableSet newCS = new HashedSet();
//               System.out.print("...and removes some more special concepts:");
               Enumeration oldCSenum = commonSupers.elements();
               while (oldCSenum.hasMoreElements()) {
                 Type oldSuper = (Type)oldCSenum.nextElement();
                 if (!thatSuper.subsumes(oldSuper)) {
                    newCS.include(oldSuper);
                 }
//                 else
//                    System.out.print(" "+oldSuper.getName());
               }
               commonSupers = newCS;
//               System.out.println();
            }
            commonSupers.include(thatSuper);
         }
//         else
//           System.out.println(" ruled out.");
      }
    }
//    System.out.println("Most general extensional supertype for "+this+" and "+that+" is "+new ConjunctiveType(commonSupers)+".");
    return new ConjunctiveType(commonSupers);
*/
  }

  public String toString() {
     return getName();
  }

   /** Compares this ConjunctiveType with another Object. Two ConjunctiveTypes
    * are considered equal iff they consist of the same set of Concepts.
    */ 
 public boolean equals(Object that) {
     if (that instanceof ConjunctiveType) {
        return equalSet(concepts,((ConjunctiveType)that).concepts);
     } else {
        return false;
     }
   }

  private static boolean equalSet(Set a, Set b) {
     if (a.size() != b.size())
        return false;
     Enumeration aEnum=a.elements();
     for (int i=a.size(); i>0; --i) {
        if (!b.includes(aEnum.nextElement()))
          return false;
     }
     return true;
  }

  /** Return an XML representation of this Type
   */
   public String toXML() {
      StringBuffer output=new StringBuffer("<TYPE>\n");
      Enumeration conceptEnum = concepts.elements();
      while (conceptEnum.hasMoreElements()) {
        Concept concept = (Concept)conceptEnum.nextElement();
        output.append(concept.toXMLref());
      }
      output.append("</TYPE>\n");
      return output.toString();
   }

}
