package fs;

import collections.*;
import java.util.*;

public class EquivRelation {

  /** Return the Unification of fs1 and fs2. */
  public static Node unify(Node fs1, Node fs2)
         throws UnificationFailure {
 // System.out.println("Unifying "+fs1.seqNo+" and "+fs2.seqNo+".");
     EquivRelation er = new EquivRelation();
     er.unify2(fs1,fs2);
     er.extensionalize();
     return er.rebuild(fs1);
  }

  /** Return the result of unifying fs2 into fs1 at the given path. */
  public static Node unify(Node fs1, String path,
                        Node fs2) throws UnificationFailure {
 // System.out.println("Unifying "+fs1.seqNo+" at "+path+" and "+fs2.seqNo+".");
     EquivRelation er = new EquivRelation();
     er.unify2(fs1.at(path),fs2);
     er.extensionalize();
     return er.rebuild(fs1);
  }

  /** Return whether fs1 and fs2 can be unified. */
  public static boolean canUnify(Node fs1, Node fs2) {
    EquivRelation er = new EquivRelation();
    try {
      er.unify2(fs1,fs2);
      return true;
    } catch (UnificationFailure uff) {
      return false;
    }
  }

  /** tie maps Nodes to Unificator Nodes. */
  private UpdatableMap tie = new HashedMap();

  /** eit maps Unificator Nodes to sets of Nodes.
    */
  private UpdatableMap eit = new HashedMap();

  /** Constructs a new Equivalence Relation. */
  public EquivRelation() {}

  /** Maps a Node to a Unificator Node.
    * To do this, the first Node has to be mapped to
    * the Unificator (using tie) and has to be included
    * in the set of members of the unificator (using eit).
    * If the first Node is itself a Unificator, all its
    * elements are also mapped to the new Unificator.
    */
  private void map(Node fs, Node uni) {
//System.out.println("mapping node "+fs.seqNo+" to node "+uni.seqNo);
     if (fs==uni)
        return;
     if (eit.includesKey(fs)) { // fs is itself an equivClass
        // map all of fs's elements to uni:
       Enumeration equivClassElems=((UpdatableSet)eit.at(fs)).elements();
       while (equivClassElems.hasMoreElements()) {
          Node memberFS = (Node)equivClassElems.nextElement();
          map(memberFS, uni);
        }
        eit.removeAt(fs);
     }

     tie.putAt(fs, uni);
     UpdatableSet equivClass;
     if (eit.includesKey(uni)) // uni has been established before
        equivClass = (UpdatableSet)eit.at(uni);
     else { // uni is a new equivClass
        equivClass = new HashedSet();
        eit.putAt(uni,equivClass);
     }
     equivClass.include(fs);
  }


  public Node getUnificator(Node fs) {
     if (tie.includesKey(fs))
        return (Node)tie.at(fs);
     else
        return fs;
  }

  /** Updates the EquivRelation and constructs Unificator Nodes
    * so that fs1 and fs2 are unified.
    * Other nodes may become unified by recursion.
    */
  public void unify2(Node fs1, Node fs2)
          throws UnificationFailure {
//System.out.println("unifying nodes "+fs1.seqNo+" and "+fs2.seqNo);
     Type unitype = fs1.getType().unify(fs2.getType());
     Node uni;
     Enumeration featenum;
     Integer featureCode;
     Enumeration concepts = unitype.concepts();
     Object firstConcept;
     if (concepts.hasMoreElements() &&
         (firstConcept=concepts.nextElement()) instanceof JavaObjectFS) {
        uni = (JavaObjectFS)firstConcept;
     } else {
       UpdatableMap features = new HashedMap();
       featenum = fs1.featureCodes();
       while (featenum.hasMoreElements()) {
          featureCode = (Integer)featenum.nextElement();
          features.putAt(featureCode,fs1.delta(featureCode));
       }
       featenum = fs2.featureCodes();
       while (featenum.hasMoreElements()) {
          featureCode = (Integer)featenum.nextElement();
          features.putAt(featureCode,fs2.delta(featureCode));
       }
       uni = new FSNode(unitype,features);
     } /* endif */

     // Update the relation:
     // both Nodes are mapped to their unificator.
     map(fs1, uni);
     map(fs2, uni);

     // search for common features of <fs1> and <fs2>:
     featenum = fs1.featureCodes();
     while (featenum.hasMoreElements()) {
        featureCode = (Integer)featenum.nextElement();
        Node fs1post = getUnificator(fs1.delta(featureCode));
        if (fs2.hasFeature(featureCode)) {
           Node fs2post = getUnificator(fs2.delta(featureCode));
           if (fs1post != fs2post) {
// System.out.println("because nodes "+fs1.seqNo+" and "+fs2.seqNo+" both have feature "
//+Feature.getFeature(featureCode)+":");
             unify2(fs1post, fs2post);
           }
        }
     }
  }

  /** Unifies all nodes in this EquivRelation according to the extensionability rule.
     */
  public void extensionalize() throws UnificationFailure {}

  /** Rebuilds a new graph which is the Unificator.
    * fs is one of the Nodes that have been unified,
    * so the algorithm knows where to start.
    */
  public Node rebuild(Node fs) {
//System.out.println("converting node "+fs.seqNo);
     Node uni;
     boolean expand;
     if (tie.includesKey(fs)) {
//System.out.println("node "+fs.seqNo+" found in relation.");
       uni = (Node)tie.at(fs);
       expand = eit.includesKey(uni);
       if (expand)
         eit.removeAt(uni);
     } else {
//System.out.print("node "+fs.seqNo+" not found in relation. Creating Copy ");
//        if (eit.includesKey(fs)) {
//System.out.println("Using unificator as new node.");
//           uni = fs;
//           eit.removeAt(uni);
//        } else {
//System.out.println("Creating copy.");
           uni = fs.duplicate();
           tie.putAt(fs, uni);
//        } /* endif */
        expand = true;
     }
//System.out.println("node "+fs.seqNo+" is now mapped to node "+uni.seqNo);
     if (expand) {
//System.out.println("Expanding node "+uni.seqNo+"...");
        Enumeration featenum = uni.featureCodes();
        while (featenum.hasMoreElements()) {
           Integer featureCode = (Integer)featenum.nextElement();
//System.out.println("Expanding feature "+Feature.getFeature(featureCode)+" of node "+uni.seqNo+"...");
           uni.updateFeature(featureCode,rebuild(uni.delta(featureCode)));
        }
     }
//else {
// System.out.println("Already expanded: "+uni.seqNo);
//}
     return uni;
  }

}

