package fs;

import collections.*;
import java.util.*;

public class FSNode extends Node {
  Type nodetype;
  static Comparator comp = new IntComparator();
  UpdatableMap feats = new RBMap(comp);

  public FSNode(Type nodetype, Map feats) throws TypeException {
    this.nodetype = nodetype;
    Enumeration features = feats.keys();
    while (features.hasMoreElements()) {
      Integer feature = (Integer)features.nextElement();
      addFeature(feature,(Node)feats.at(feature));
    }
  }

  public FSNode(Type nodetype) {
    this.nodetype = nodetype;
  }

  public FSNode() {
    this(Type.ANY);
  }

  public FSNode(Feature feat) throws TypeException { this(new Feature[]{feat}); }

  public FSNode(Type nodetype, Feature feat) throws TypeException {
    this(nodetype, new Feature[]{feat});
  }

  public FSNode(Feature[] features) throws TypeException {
    this(Type.TOP, features);
  }

  public FSNode(Type nodetype, Feature[] features) throws TypeException {
     this(nodetype);
     for (int i=0; i < features.length; ++i)
        addFeature(features[i].feature, features[i].value);
  }

  public FSNode(String nodetype) throws UnificationFailure {
    this(ConjunctiveType.getType(nodetype));
  }

  public FSNode(String nodetype, UpdatableMap feats) throws UnificationFailure, TypeException {
    this(ConjunctiveType.getType(nodetype),feats);
  }

  FSNode(Type nodetype, UpdatableMap feats) {
    this.nodetype = nodetype;
    this.feats =  feats;
  }

  public Type getType() { return nodetype; }

  public void refine(Type type) throws UnificationFailure {
     nodetype = nodetype.unify(type);
  }

  public CollectionEnumeration featureCodes() {
      LinkedList ll=new LinkedList();
      ll.appendElements(feats.keys());
      return ll.elements();
  }

  public boolean hasFeature(Integer featureCode) {
     return feats.includesKey(featureCode);
  }

  public Node delta(Integer featureCode) throws NoSuchElementException {
    return (Node)feats.at(featureCode);
  }

  /** Updates the value of an existing feature.
    *  This method should only be called during construction of
    *  a FSNode and should not be used to create new features.
    *  Use addFeature() instead!
    */
  void updateFeature(Integer featureCode, Node newValue) {
     feats.putAt(featureCode, newValue);
  }

  /** Adds a new feature with the given name and value.
    *  This method should only be called during construction of
    *  a FSNode and should not be used to change existing features.
    *  Use updateFeature() instead!
    */
  public FSNode addFeature(String feature, Node value) throws TypeException {
    return addFeature(Feature.getCode(feature),value);
  }

  /** Adds a new feature with the given name (as feature code) and value.
    *  This method should only be called during construction of
    *  a FSNode and should not be used to change existing features.
    *  Use updateFeature() instead!
    */
  public FSNode addFeature(Integer featureCode, Node value) throws TypeException {
     if (hasFeature(featureCode)) {
       throw new RuntimeException("Feature "+Feature.getFeature(featureCode)+" redefined!");
     }
     if (nodetype.isApprop(featureCode) && nodetype.appropType(featureCode).subsumes(value.getType())) {
       feats.putAt(featureCode, value);
       return this;
     } else {
        throw new TypeException();
     }
   }

  public Node duplicate() {
     return new FSNode(nodetype, (UpdatableMap)feats.duplicate());
  }

  private String listToString(TagMap map, String indent) {
    StringBuffer output = new StringBuffer();
    if (Type.NELIST.subsumes(nodetype)) {
	  if (hasFeature(Feature.HEAD))
	      output = output.append(delta(Feature.HEAD).toString(map,indent+" "))
		.append(indent);
	  else
	    output = output.append("(no head feature!)");
	  if (hasFeature(Feature.TAIL)) {
	    FSNode tailfs = (FSNode)delta(Feature.TAIL);
	    int tag = map.getTag(tailfs);
	    if (tag != 0) {
	      output = output.append("| ")
		.append(tailfs.toString(map,indent+"   "));
	    } else {
	      output = output.append(tailfs.listToString(map,indent));
	    }
	  } else
	    output = output.append("(no tail feature!)");
    } else if (!Type.ELIST.equals(nodetype)) {
      output.append("| ").append(toString(map,indent+"   "));
    }
    return output.toString();
  }

  private String innerToString(TagMap map, String indent) {
        StringBuffer output = new StringBuffer();
	String whitespace = "";
	if (!nodetype.equals(Type.TOP)) {
	  output = output.append(nodetype.getName());
	  whitespace = indent;
	}
	CollectionEnumeration features = featureCodes();
	while (features.hasMoreElements()) {
	  Integer featureCode = (Integer)features.nextElement();
	  String feature = Feature.getFeature(featureCode);
	  output = output.append(whitespace).append(feature).append(": ")
	    .append(delta(featureCode)
		    .toString(map,indent+FeatureStructure.indent(3+feature.length())));
	  whitespace = indent;
	}
    return output.toString();
  }

  protected String untaggedToString(TagMap map, String indent) {
     StringBuffer output = new StringBuffer();
     if (Type.NELIST.subsumes(nodetype) || Type.ELIST.equals(nodetype)) {
	  output = output.append("<").append(listToString(map,indent))
	    .append(">");
     } else {
          output.append("[").append(innerToString(map,indent)).append("]");
     }
     return output.toString();
  }

}
