package fs;

import collections.*;
import java.util.NoSuchElementException;

/** The class FeatureStructure defines all methods that are needed
  * to construct, access and unify Feature Structures.
  * A FeatureStructure mainly wraps a Node and provides
  * customized access to the Node's type, features, etc.
  */
public class FeatureStructure {

  private static int counter=0;
  public final int seqNo=++counter;

  private Node node;
  protected int hashCode; // as the hashCode is a finicky computation, it is being cached

  public FeatureStructure(Node node) {
     this.node = node;
     node.fs = this;
     // compute hashCode:
     hashCode = node.getType().hashCode();
     // features contribute to hashCode:
     CollectionEnumeration featenum=node.featureCodes();
     while (featenum.hasMoreElements()) {
        hashCode+=featenum.nextElement().hashCode();
     }
     /* feature values could also contribute to the hashCode,
      * but this is
      * difficult because there might be cycles and
      * the hashcode should not depend on the order
      * in which the nodes are constructed.
      */
  }

  /** Get the Type of the Feature Structure. */
  public Type getType() {
     return node.getType();
  }

  /** Get the root node of the Feature Structure. */
  public Node getRoot() {
     return node;
  }

  /** Get the Java Object contained in this Feature Structure.
     * Throws ClassCastException (?) if the root node is not
     * a JavaObjectFS.
     */
  public Object getJavaObject() {
     return ((JavaObjectFS)node).getJavaObject();
  }

  /** Get the Type of the Feature Structure found at the given path.
    *  A path is a sequence of Features (meant to be starting at the root node).
    *  The syntax of a path is feature_1:feature_2:...feature_n.
    */
  public Type getType(String path) throws NoSuchElementException {
     return node.at(path).getType();
  }

  /** Test whether this Feature Structure subsumes that Feature Structure. */
  public boolean subsumes(FeatureStructure that) {
    return node.subsumes(new HashedMap(), that.node);
  }

  /** Test whether this Feature Structure is subsumed by that Feature Structure. */
  public boolean subsumedBy(FeatureStructure that) {
     return that.subsumes(this);
  }

  /** Calculate the unification of this Feature Structure and that Feature Structure. */
  public FeatureStructure unify(FeatureStructure that) throws UnificationFailure {
     return new FeatureStructure(EquivRelation.unify(this.node,that.node));
  }

  /** Test whether this Feature Structure can be unified with that Feature Structure. */
  public boolean canUnify(FeatureStructure that) {
     return EquivRelation.canUnify(this.node,that.node);
  }

  /** Return an Enumeration of the codes of all Features of this Feature Structure. */
  public CollectionEnumeration featureCodes() {
     return node.featureCodes();
  }

  /** Returns whether this Feature Structure has the given Feature. */
  public boolean hasFeature(String feature) {
     return node.hasFeature(Feature.getCode(feature));
  }

  /** Returns whether this Feature Structure has the Feature given by its code. */
  public boolean hasFeature(Integer featureCode) {
     return node.hasFeature(featureCode);
  }

  /** Returns the Feature Structure at the given path.
    *  A path is a sequence of Features (meant to be starting at the root node).
    *  The syntax of a path is feature_1:feature_2:...feature_n.
    *  The empty String returns the Feature Structure itself.
    */
  public FeatureStructure at(String path) throws NoSuchElementException {
     return node.at(path).asFeatureStructure();
  }

  public FeatureStructure change(String feature, FeatureStructure newValue) {
    Integer featureCode = Feature.getCode(feature);
    FSNode newNode = (FSNode)node.duplicate();
    newNode.feats.putAt(featureCode,newValue.node);
    return newNode.asFeatureStructure();
  }


  public boolean equals(Object that) {
    if (this == that)
      return true;
    if (that instanceof FeatureStructure)
      return this.subsumes((FeatureStructure)that) && ((FeatureStructure)that).subsumes(this);
    else
      return false;
  }

  public int hashCode() {
     return hashCode;
  }

  public String toString() {
    TagMap map = new TagMap(node);
    return "\n" + node.toString(map,"\n ") + "\n";
  }

  public final static String indent(int depth) {
    StringBuffer space = new StringBuffer(depth+1);
    for (int i = 0; i < depth; ++i)
      space.append(' ');
    return space.toString();
  }

}


