package fs;

import collections.*;
import java.util.*;

public class JavaConcept implements Concept {

  public static final Set emptySet = new HashedSet();
  static private UpdatableMap javaConceptsByName = new HashedMap();
  private Class javaClass;
  private static String imported[] = {"","java.lang."};

  static {
    new JavaConcept(Boolean.TYPE);
    new JavaConcept(Byte.TYPE);
    new JavaConcept(Character.TYPE);
    new JavaConcept(Short.TYPE);
    new JavaConcept(Integer.TYPE);
    new JavaConcept(Long.TYPE);
    new JavaConcept(Float.TYPE);
    new JavaConcept(Double.TYPE);
  }

  /** Construct a new wrapper for the given Java Type. */
  private JavaConcept(Class javaClass) {
    this.javaClass = javaClass;
    javaConceptsByName.putAt(javaClass.getName(),this);
  }

  public static JavaConcept getJavaConcept(Class javaClass) {
    String name=javaClass.getName();
    if (javaConceptsByName.includesKey(name))
       return (JavaConcept)javaConceptsByName.at(name);
    else
       return new JavaConcept(javaClass);
  }

  /** Find a wrapper for the Java Type given by its name. */
  public static JavaConcept getJavaConcept(String javaClassName)
           throws TypeException {
    if (javaConceptsByName.includesKey(javaClassName))
       return (JavaConcept)javaConceptsByName.at(javaClassName);
    for (int i=0; i<imported.length; ++i) {
       try {
          return new JavaConcept(Class.forName(imported[i]+javaClassName));
       } catch (ClassNotFoundException cnf1) {}
    }
    throw new TypeException();
  }

  /** Return the name of this Type. */
  public String getName() {
    return javaClass.getName();
  }

 /** Return the Java Type that is wrapped by this JavaConcept.
   */
  public Class getJavaClass() {
     return javaClass;
  }

  /** Return whether this Type is extensional. */
  public boolean isExtensional() {
    return false; // JavaConcepts are not extensional.
  }

  /** Return whether the feature <feature> is appropriate in this Type. */
  public boolean isApprop(Integer featureCode) {
    return false; // A  JavaConcept has no features.
  }

  /** Return the required Type of the Value found under the given feature.
    * The feature has to be appropriate for this Type.
    */
  public Type appropType(Integer featureCode) throws NoSuchElementException {
    throw new NoSuchElementException(); // A JavaConcept has no features.
  }

  /** Return an Enumeration of all appropriate features. */
  public CollectionEnumeration appropFeatureCodes() {
    return emptySet.elements(); // A JavaConcept has no features.
  }

  /** Return whether this JavaConcept is-a <that> Concept.
    * In other words, return wheter this Concept is more special than <that>
    * Concept.
    */
  public boolean isa(Concept that) {
    if (that instanceof JavaConcept) {
        Class thatJavaClass = ((JavaConcept)that).getJavaClass();
        return thatJavaClass.isAssignableFrom(javaClass);
    }
    return false;
  }

  /** Return whether this JavaConcept is-not-a "that" Concept.
    * In other words, return wheter this JavaConcept is incompatible with 
    *  "that" JavaConcept. This is the case if
    * <ol><i>the other Concept is not a JavaConcept,
    *        <i>both JavaConcepts are classes which are not in a subclass 
    *              relation, or</i>
    *        <i> the Java classes/interfaces define methods in an 
    *               incompatible way.</i></ol>
    */
   public boolean isNotA(Concept that) {
    if (this.isa(that) || that.isa(this))
       return false;
    if (that instanceof JavaObjectFS) {
       return true; // the only chance was: that.isa(this)
    }
    if (javaClass.isInterface())
       return false; // method-inconsistencies are not checked yet!
    if (that instanceof JavaConcept) {
        Class thatJavaClass = ((JavaConcept)that).getJavaClass();
        if (thatJavaClass.isInterface())
           return false; // method-inconsistencies are not checked yet!
    }
    return true;
  }

  /** Look for the most general common extensional supertype of this and "that". */
  public CollectionEnumeration extensionalSuperconcepts() {
     return null; // not extensional, should never be called!
  }

  public void remove() {
     javaConceptsByName.removeAt(javaClass.getName());
  }


  /** Return an XML representation of this Concept.
   */
   public String toXML() {
      return "<JAVACONCEPT class=\""+getName()+"\"/>";
   }

  /** Return an XML representation of a reference to this Concept.
   */
  public String toXMLref() {
      return toXML();
  }

}
