package fs;

import collections.*;
import java.util.*;
import de.renew.function.Value;

public class JavaObjectFS extends Node implements Concept {

  private static final Set empty = new HashedSet();
  private Object javaObject;
  private JavaConcept concept;
  private Type type;

  /** Construct a new wrapper for the given Java Object. */
  public JavaObjectFS(Object javaObject) {
    this.javaObject = javaObject;
    Class clazz;
    if (javaObject instanceof Value) {
        clazz=de.renew.util.Types.typify(((Value)javaObject).value.getClass());
    } else {
        clazz=javaObject.getClass();
    }
    concept = JavaConcept.getJavaConcept(clazz);
    type = new ConjunctiveType(this);
  }

  /** Return the Java Object that is wrapped by this JavaObjectFS.
   */
  public Object getJavaObject() {
     return javaObject;
  }

  public int hashCode() { return javaObject.hashCode(); }

 public boolean equals(Object that) {
     if (that instanceof JavaObjectFS) {
        return javaObject.equals(((JavaObjectFS)that).javaObject);
     } else {
        return javaObject.equals(that);
     } /* endif */
  }

 /* Implementation of Node methods: */

  /** Return an Enumeration of the codes of all Features of this Feature Structure. */
  public CollectionEnumeration featureCodes() {
     return empty.elements();  // A JavaObjectFS has no features.
  }

  /** Returns whether this Feature Structure has the Feature given by its code. */
  public boolean hasFeature(Integer featureCode) {
     return false;  // A JavaObjectFS has no features.
  }

  /** Returns the Feature Structure at the Feature given by its code. */
  public Node delta(Integer featureCode) throws NoSuchElementException {
     throw new NoSuchElementException(); // A JavaObjectFS has no features.
  }

  public Node at(String path) throws NoSuchElementException {
     throw new NoSuchElementException(); // A JavaObjectFS has no features.
  }

  void updateFeature(Integer featureCode, Node newValue) { }

  protected String untaggedToString(TagMap map, String indent) {
     return "(" + getName() + ")";
  }



  /* Implementation of Concept methods: */

  /** Return the name of this Concept. */
  public String getName() {
     if (javaObject instanceof String)
        return '"' + javaObject.toString() + '"';
     else 
        return javaObject.toString();
  }

  /** Return whether this Concept is-a <that> Concept.
    * In other words, return wheter this Concept is more special than <that>
    * Concept.
    */
  public boolean isa(Concept that) {
    if (that instanceof JavaObjectFS) {
       Object thatJavaObject = ((JavaObjectFS)that).getJavaObject();
       return javaObject.equals(thatJavaObject);
    } else if (that instanceof Bound) {
       return ((Bound)that).subsumesObject(javaObject);
    } else
        return concept.isa(that);
  }

  /** Return whether this Concept is-not-a <that> Concept.
    * In other words, return wheter this Concept is incompatible with <that>
    * Concept.
    */
  public boolean isNotA(Concept that) {
    return !isa(that);
  }

  public CollectionEnumeration extensionalSuperconcepts() {
     return null;
  }

  public void remove() { }

  /** Get the Type of the Feature Structure. */
  public Type getType() {
     return type;
  }

  Node duplicate() {
     return this; // A JavaObjectFS is extensional and thus need not be copied!
  }

  /** Return whether this Concept is extensional. */
  public boolean isExtensional() {
    return true; // by default, all JavaObjectFSs are extensional.
  }

  /** Return whether the feature <feature> is appropriate for this Concept. */
  public boolean isApprop(Integer featureCode) {
    return false; // A JavaObjectFS has no features.
  }

  /** Return the required Type of the Value found under the given feature.
    * The feature has to be appropriate for this Type.
    */
  public Type appropType(Integer featureCode) throws NoSuchElementException {
    throw new NoSuchElementException(); // A JavaObjectFS has no features.
  }

  /** Return an Enumeration of all appropriate features. */
  public CollectionEnumeration appropFeatureCodes() {
    return empty.elements(); // A JavaObjectFS has no features.
  }

  /** Return an XML representation of this Concept.
   */
   public String toXML() {
      return "<JAVAOBJECT/>";
   }

  /** Return an XML representation of a reference to this Concept.
   */
  public String toXMLref() {
      return "<JAVAOBJECTREF/>";
  }

}
