package fs;

import collections.*;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;

/** The abstract class Node defines all methods that are needed
  * to construct, access and unify Feature Structures Nodes.
  * A Node has a Type and a number of Features that as a value
  * contain another Node of a Type appropriate for that Feature.
  */
public abstract class Node {

  private static int counter=0;
  public final int seqNo=++counter;

  FeatureStructure fs=null;

  public FeatureStructure asFeatureStructure() {
     if (fs==null)
        fs=new FeatureStructure(this);
     return fs;
  }

  /** Get the Type of the Node. */
  public abstract Type getType();

  public boolean subsumes(UpdatableMap h, Node that) {
    if (h.includesKey(this))
      return h.at(this) == that;
    if (!getType().subsumes(that.getType()))
      return false;
    CollectionEnumeration features = this.featureCodes();
    h.putAt(this,that);
    while (features.hasMoreElements()) {
      Integer featureCode = (Integer)features.nextElement();
      Node thispost = delta(featureCode);
      if (that.hasFeature(featureCode)) {
	Node thatpost = that.delta(featureCode);
	if (!thispost.subsumes(h, thatpost))
	  return false;
      } else
	return false;
    }
    return true;
  }

  /** Return an Enumeration of the codes of all Features of this Feature Structure. */
  public abstract CollectionEnumeration featureCodes();

  /** Returns whether this Feature Structure has the Feature given by its code. */
  public abstract boolean hasFeature(Integer featureCode);

  /** Returns the Feature Structure at the Feature given by its code. */
  public abstract Node delta(Integer featureCode) throws NoSuchElementException;

  public Node at(String path) throws NoSuchElementException {
     StringTokenizer pathenum=new StringTokenizer(path,":");
     Node curr=this;
     while (pathenum.hasMoreTokens()) {
        curr=curr.delta(Feature.getCode(pathenum.nextToken()));
     }
     return curr;
  }

  protected String toString(TagMap map, String indent) {
    StringBuffer output = new StringBuffer();
//output.append('*').append(seqNo); indent += FeatureStructure.indent(output.length());
    int tag = map.getTag(this);
    if (tag != 0) {
      output = output.append("#").append(Math.abs(tag));
      indent += FeatureStructure.indent(output.length());
    }
    if (!map.isVisited(this)) {
      map.setVisited(this);
      output.append(untaggedToString(map,indent));
    }
    return output.toString();
  }

  abstract protected String untaggedToString(TagMap map, String indent);

  abstract Node duplicate();

  abstract void updateFeature(Integer featureCode, Node newValue);

}

