package fs;

import collections.*;
import java.util.*;

public class TagMap {

   private UpdatableMap tagmap=new HashedMap();
   private UpdatableSet visited;
   private Node root;

   /** Builds a tag map for a Feature Structure.
      * The tag map provides an Integer for every node that is connected
      * with the given root node. If a node is labelled zero, it has exactly
      * one predescessor and need not be tagged. All other nodes have
      * multiple predecessors and get assigned a unique Integer that can
      * be used as a tag number.
      */
   public TagMap(Node node) {
        root=node;
        visited=new HashedSet();
        buildTagMap(root,0);
   }

   public Node getRoot() { return root; }

   private int buildTagMap(Node node, int tagcounter) {
    int tag;
    if (tagmap.includesKey(node)) {
      tag = ((Integer)tagmap.at(node)).intValue();
      if (tag == 0)     // visited before
	tagmap.putAt(node,new Integer(++tagcounter)); // give node a name
    } else {
      tagmap.putAt(node,new Integer(0)); // mark as visited
      // do this recursively for all nodes:
      Enumeration features = node.featureCodes();
      while (features.hasMoreElements()) {
	Integer featureCode = (Integer)features.nextElement();
	tagcounter = buildTagMap(node.delta(featureCode), tagcounter);
      }
    }
    return tagcounter;
  }

  public void resetVisited() {
     visited=new HashedSet();
  }

  public final static String indent(int depth) {
    StringBuffer space = new StringBuffer(depth+1);
    for (int i = 0; i < depth; ++i)
      space.append(' ');
    return space.toString();
  }

  public int getTag(Node node) {
    if (tagmap.includesKey(node))
      return ((Integer)tagmap.at(node)).intValue();
   else
      return 0;
  }

  public boolean isVisited(Node node) {
    return visited.includes(node);
  }

  public void setVisited(Node node) {
     visited.include(node);
  }

}
